# Copyright (c) 2025 GeoUtils developers
# Copyright (c) 2025 Centre National d'Etudes Spatiales (CNES)
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Filters class to remove outliers and reduce noise in rasters.
"""

from __future__ import annotations

from collections.abc import Callable
from typing import Any, Literal

import numpy as np
import scipy
from packaging.version import Version
from scipy.ndimage import uniform_filter

from geoutils._misc import import_optional
from geoutils._typing import NDArrayNum

if Version(scipy.__version__) > Version("1.16.0"):
    generic_filter_scipy = scipy.ndimage.vectorized_filter
    _has_vectorized_filter = True
else:
    generic_filter_scipy = scipy.ndimage.generic_filter
    _has_vectorized_filter = False

try:
    from numba import jit, prange

    _HAS_NUMBA = True
except ImportError:
    _HAS_NUMBA = False

    def jit(*args: Any, **kwargs: Any) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        """
        Fake jit decorator if numba is not installed
        """

        def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
            return func

        return decorator


def _filter(array: NDArrayNum, method: str | Callable[..., NDArrayNum], size: int = 3, **kwargs: Any) -> NDArrayNum:
    """
    Dispatch filter application by method name or custom callable.

    :param array: Array to filter.
    :param method: Filter method name or callable.
    """
    filter_map: dict[str, Callable[..., NDArrayNum]] = {
        "gaussian": gaussian_filter,
        "median": lambda arr, size=size, **_: generic_filter_scipy(arr, np.nanmedian, size=size),
        "mean": lambda arr, size=size, **_: generic_filter_scipy(arr, np.nanmean, size=size),
        "max": lambda arr, size=size, **_: generic_filter_scipy(arr, np.nanmax, size=size),
        "min": lambda arr, size=size, **_: generic_filter_scipy(arr, np.nanmin, size=size),
        "distance": distance_filter,
    }

    if isinstance(method, str):
        if method not in filter_map:
            raise ValueError(f"Unsupported filter method '{method}'. Available: {list(filter_map)}")
        func = filter_map[method]
    elif callable(method):
        func = method
    else:
        raise TypeError("`method` must be a string or a callable.")

    return func(array, **kwargs)


def gaussian_filter(array: NDArrayNum, sigma: float, **kwargs: Any) -> NDArrayNum:
    """
    Apply a Gaussian filter to a raster that may contain NaNs.
    N.B: kernel_size is set automatically based on sigma.

    :param array: The input array to be filtered.
    :param sigma: The sigma of the Gaussian kernel

    :returns: The filtered array (same shape as input)
    """

    if array.ndim == 1:
        raise ValueError("Gaussian filter can't be applied to 1D arrays.")

    # Boolean mask: True where NaN
    mask = np.isnan(array)
    mask_f = (~mask).astype(float)

    # Replace NaNs with 0
    arr_filled = np.where(mask, 0, array)

    # Apply gaussian filter to values and mask
    filtered = scipy.ndimage.gaussian_filter(arr_filled, sigma, **kwargs)
    normalization = scipy.ndimage.gaussian_filter(mask_f, sigma, **kwargs)

    # Avoid division by zero
    with np.errstate(invalid="ignore", divide="ignore"):
        filtered /= normalization

    # Where normalization is zero, set result to NaN
    filtered[normalization == 0] = np.nan

    return filtered


@jit(nopython=True, parallel=True)  # type: ignore
def median_filter_numba(array: NDArrayNum, size: int) -> NDArrayNum:
    """
    Apply a median filter to a raster that may contain NaNs, using numbas's implementation.

    :param array: The input array to be filtered.
    :param size: The size of the window to use (must be odd).

    :returns: The filtered array (same shape as input).
    """

    if size % 2 == 0:
        raise ValueError("`size` must be odd.")

    N1, N2 = array.shape
    pad = size // 2

    padded = np.full((N1 + 2 * pad, N2 + 2 * pad), np.nan, dtype=array.dtype)

    for row in range(N1):
        for col in range(N2):
            padded[row + pad, col + pad] = array[row, col]

    outputs = np.full((N1, N2), np.nan, dtype=array.dtype)

    for row in prange(N1):
        for col in prange(N2):
            window = padded[row : row + size, col : col + size]
            outputs[row, col] = np.nanmedian(window)

    return outputs


def median_filter(array: NDArrayNum, size: int, engine: Literal["scipy", "numba"] = "scipy") -> NDArrayNum:
    """
    Apply a median filter to a raster that may contain NaNs.

    For 2D arrays, the filter is applied over both dimensions.
    For 3D arrays, the filter is applied independently to each 2D slice
    (i.e., only along the horizontal dimensions, not across the third dimension).

    This differs from scipy's built-in median_filter, which applies
    the filter across all dimensions by default.

    :param array: The input array to be filtered.
    :param size: The size of the filtering window (must be odd).
    :param engine: Filtering engine to use, either "scipy" or "numba".
    :returns: The filtered array (same shape as input).
    """

    if size % 2 == 0:
        raise ValueError("`size` must be odd.")

    if array.ndim == 2:
        return _apply_median_filter_2d(array, size, engine)
    elif array.ndim == 3:
        return np.stack([_apply_median_filter_2d(slice_, size, engine) for slice_ in array])
    raise ValueError("Input array must be 2D or 3D.")


def _apply_median_filter_2d(array: NDArrayNum, size: int, engine: Literal["scipy", "numba"]) -> NDArrayNum:
    """
    Apply a 2D median filter on an array that may contain NaNs.

    :param array: 2D input array to filter, may contain NaNs.
    :param size: Size of the median filter window (must be odd).
    :param engine: Filtering engine to use, either "scipy" or "numba".
    :returns: Filtered array of the same shape as input.
    """

    nans = np.isnan(array)

    if engine == "scipy":
        median_vals = generic_filter_scipy(array, np.nanmedian, size=size, mode="constant", cval=np.nan)
        return np.where(nans, array, median_vals)

    else:
        if not _HAS_NUMBA:
            # Call import simply to raise an error consistent with other optional imports
            import_optional("numba")

        median_vals = median_filter_numba(array, size)
        return np.where(nans, array, median_vals)


def mean_filter(array: NDArrayNum, size: int) -> NDArrayNum:
    """
    Apply a mean filter to a 2D array that may contain NaNs.

    :param array: 2D input array
    :param size: size of the square kernel
    :no_data: no data value
    :return: filtered array with same shape
    """
    if array.ndim != 2:
        raise ValueError(f"Invalid array shape {array.shape}, expected 2D.")

    # Mask nodata values
    nans = np.isnan(array)
    mask = ~np.isnan(array)
    array_filled = np.where(mask, array, 0)
    # Compute sum over the kernel
    sum_vals = uniform_filter(array_filled, size=size, mode="constant", cval=0.0)
    # Count of valid (non-nodata) pixels in the kernel
    count_vals = uniform_filter(mask.astype(float), size=size, mode="constant", cval=0.0)

    with np.errstate(invalid="ignore", divide="ignore"):
        mean_vals = sum_vals / count_vals

    return np.where(nans, array, mean_vals)


def min_filter(array: NDArrayNum, size: int = 5, **kwargs: Any) -> NDArrayNum:
    """
    Apply a minimum filter to a raster that may contain NaNs, using scipy's implementation.

    :param array: The input array to be filtered.
    :param size:  the shape that is taken from the input array, at every element position,
    to define the input to the filter function

    :returns: The filtered array (same shape as input).
    """
    # Check that array dimension is 2 or 3
    if array.ndim not in [2, 3]:
        raise ValueError(f"Invalid array shape given: {array.shape}. Expected 2D or 3D array.")

    nans = np.isnan(array)
    # We replace temporarily NaNs by infinite values during filtering to avoid spreading NaNs
    array_nans_replaced = np.where(nans, np.inf, array)
    array_nans_replaced_f = scipy.ndimage.minimum_filter(
        array_nans_replaced, size=size, mode="constant", cval=np.inf, **kwargs
    )
    # In the end, we want the filtered array without infinite values, so we put back NaNs
    return np.where(nans, array, array_nans_replaced_f)


def max_filter(array: NDArrayNum, size: int = 5, **kwargs: Any) -> NDArrayNum:
    """
    Apply a maximum filter to a raster that may contain NaNs, using scipy's implementation.

    :param array: the input array to be filtered.
    :param size:  the shape that is taken from the input array, at every element position,
    to define the input to the filter function

    :returns: the filtered array (same shape as input).
    """
    # Check that array dimension is 2 or 3
    if array.ndim not in [2, 3]:
        raise ValueError(f"Invalid array shape given: {array.shape}. Expected 2D or 3D array.")

    nans = np.isnan(array)
    # We replace temporarily NaNs by negative infinite values during filtering to avoid spreading NaNs
    array_nans_replaced = np.where(nans, -np.inf, array)
    array_nans_replaced_f = scipy.ndimage.maximum_filter(
        array_nans_replaced, size=size, mode="constant", cval=-np.inf, **kwargs
    )
    # In the end we want the filtered array without negative infinite values, so we put back NaNs
    return np.where(nans, array, array_nans_replaced_f)


def distance_filter(array: NDArrayNum, radius: float, outlier_threshold: float) -> NDArrayNum:
    """
    Filter out pixels whose value is distant more than a set threshold from the average value of all neighbor \
    pixels within a given radius.
    Filtered pixels are set to NaN.
    For npw, we use the gaussian filter for calculated the average value

    :param array: the input array to be filtered.
    :param radius: the radius in which the average value is calculated (for Gaussian filter, this is sigma).
    :param outlier_threshold: the minimum difference abs(array - mean) for a pixel to be considered an outlier.

    :returns: the filtered array (same shape as input)
    """
    # Create mask of valid (finite) values
    valid_mask = np.isfinite(array)

    # Smooth both the data and the valid mask
    smoothed = gaussian_filter(np.nan_to_num(array, nan=0.0), sigma=radius)
    normalization = gaussian_filter(valid_mask.astype(float), sigma=radius)

    # Avoid division by zero
    with np.errstate(invalid="ignore", divide="ignore"):
        local_mean = smoothed / normalization

    # Compute the outliers
    diff = np.abs(array - local_mean)
    outliers = (diff > outlier_threshold) & valid_mask

    # Create output with outliers set to NaN
    out_array = array.copy()
    out_array[outliers] = np.nan

    return out_array


def generic_filter(
    array: NDArrayNum,
    filter_function: Callable[..., NDArrayNum],
    **kwargs: Any,
) -> NDArrayNum:
    """
    Apply a filter from a function.

    :param array: the input array to be filtered.
    :param filter_function: the function of the filter.

    :returns: the filtered array (same shape as input).
    """
    # Check that array dimension is 2 or 3
    if array.ndim not in [2, 3]:
        raise ValueError(f"Invalid array shape given: {array.shape}. Expected 2D or 3D array.")
    return filter_function(array, **kwargs)



"""Tests for geotransformations of vectors."""

from __future__ import annotations

import re

import numpy as np
import pytest
from geopandas.testing import assert_geodataframe_equal, assert_geoseries_equal

import geoutils as gu


class TestGeotransformations:

    landsat_b4_path = gu.examples.get_path_test("everest_landsat_b4")
    landsat_b4_crop_path = gu.examples.get_path_test("everest_landsat_b4_cropped")
    everest_outlines_path = gu.examples.get_path_test("everest_rgi_outlines")
    aster_dem_path = gu.examples.get_path_test("exploradores_aster_dem")
    aster_outlines_path = gu.examples.get_path_test("exploradores_rgi_outlines")

    def test_reproject(self) -> None:
        """Test that the reproject function works as intended"""

        v0 = gu.Vector(self.aster_outlines_path)
        r0 = gu.Raster(self.aster_dem_path)
        v1 = gu.Vector(self.everest_outlines_path)

        # First, test with a EPSG integer
        v1 = v0.reproject(crs=32617)
        assert isinstance(v1, gu.Vector)
        assert v1.crs.to_epsg() == 32617

        # Check the inplace behaviour matches the not-inplace one
        v2 = v0.copy()
        v2.reproject(crs=32617, inplace=True)
        v2.vector_equal(v1)

        # Check that the reprojection is the same as with geopandas
        gpd1 = v0.ds.to_crs(epsg=32617)
        assert_geodataframe_equal(gpd1, v1.ds)

        # Second, with a Raster object
        v2 = v0.reproject(r0)
        assert v2.crs == r0.crs

        # Third, with a Vector object that has a different CRS
        assert v0.crs != v1.crs
        v3 = v0.reproject(v1)
        assert v3.crs == v1.crs

        # Fourth, check that errors are raised when appropriate
        # When no destination CRS is defined, or both dst_crs and dst_ref are passed
        with pytest.raises(ValueError, match=re.escape("Either of `ref` or `crs` must be set. Not both.")):
            v0.reproject()
            v0.reproject(ref=r0, crs=32617)
        # If input of wrong type
        with pytest.raises(TypeError, match=re.escape("Type of ref must be a raster or vector.")):
            v0.reproject(ref=10)  # type: ignore

    test_data = [[landsat_b4_path, everest_outlines_path], [aster_dem_path, aster_outlines_path]]

    @pytest.mark.parametrize("data", test_data)  # type: ignore
    def test_crop(self, data: list[str]) -> None:
        # Load data
        raster_path, outlines_path = data
        rst = gu.Raster(raster_path)
        outlines = gu.Vector(outlines_path)

        # Need to reproject to r.crs. Otherwise, crop will work but will be approximate
        # Because outlines might be warped in a different crs
        outlines.ds = outlines.ds.to_crs(rst.crs)

        # Crop
        outlines_new = outlines.copy()
        outlines_new.crop(crop_geom=rst, inplace=True)

        # Check default behaviour - crop and return copy
        outlines_copy = outlines.crop(crop_geom=rst)

        # Crop by passing bounds
        outlines_new_bounds = outlines.copy()
        outlines_new_bounds.crop(crop_geom=list(rst.bounds), inplace=True)
        assert_geodataframe_equal(outlines_new.ds, outlines_new_bounds.ds)
        # Check the return-by-copy as well
        assert_geodataframe_equal(outlines_copy.ds, outlines_new_bounds.ds)

        # Verify that geometries intersect with raster bound
        rst_poly = gu.projtools.bounds2poly(rst.bounds)
        intersects_new = []
        for poly in outlines_new.ds.geometry:
            intersects_new.append(poly.intersects(rst_poly))

        assert np.all(intersects_new)

        # Check that some of the original outlines did not intersect and were removed
        intersects_old = []
        for poly in outlines.ds.geometry:
            intersects_old.append(poly.intersects(rst_poly))

        assert np.sum(intersects_old) == np.sum(intersects_new)

        # Check that some features were indeed removed if any geometry didn't intersect the raster bounds
        if any(~np.array(intersects_old)):
            assert np.sum(~np.array(intersects_old)) > 0

        # Check that error is raised when cropGeom argument is invalid
        with pytest.raises(TypeError, match="Crop geometry must be a Raster, Vector, or list of coordinates."):
            outlines.crop(1, inplace=True)  # type: ignore

    def test_translate(self) -> None:

        vector = gu.Vector(self.everest_outlines_path)

        # Check default behaviour is not inplace
        vector_shifted = vector.translate(xoff=2.5, yoff=5.7)
        assert isinstance(vector_shifted, gu.Vector)
        assert_geoseries_equal(vector_shifted.geometry, vector.geometry.translate(xoff=2.5, yoff=5.7))

        # Check inplace behaviour works correctly
        vector2 = vector.copy()
        output = vector2.translate(xoff=2.5, yoff=5.7, inplace=True)
        assert output is None
        assert_geoseries_equal(vector2.geometry, vector_shifted.geometry)



# Copyright (c) 2025 GeoUtils developers
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Setup of runtime-compile configuration of GeoUtils."""

from __future__ import annotations

import configparser
import os
from typing import Any

# The setup is inspired by that of Matplotlib and Geowombat
# https://github.com/matplotlib/matplotlib/blob/main/lib/matplotlib/rcsetup.py
# https://github.com/jgrss/geowombat/blob/main/src/geowombat/config.py

_config_ini_file = os.path.abspath(os.path.join(os.path.dirname(__file__), "config.ini"))

# Validators: to check the format of user inputs


def validate_bool(b: bool | str | int) -> bool:
    """Convert b to ``bool`` or raise."""
    if isinstance(b, str):
        b = b.lower()
    if b in ("t", "y", "yes", "on", "true", "1", 1, True):
        return True
    elif b in ("f", "n", "no", "off", "false", "0", 0, False):
        return False
    else:
        raise ValueError(f"Cannot convert {b!r} to bool")


# Map the parameter names with a validating function to check user input
_validators = {
    "shift_area_or_point": validate_bool,
    "warn_area_or_point": validate_bool,
}


class GeoUtilsConfigDict(dict):  # type: ignore
    """Class for a GeoUtils config dictionary"""

    def __setitem__(self, k: str, v: Any) -> None:
        """We override setitem to check user input."""

        validate_func = _validators[k]
        new_value = validate_func(v)
        super().__setitem__(k, new_value)

    def _set_defaults(self, path_init_file: str) -> None:
        """A function to set"""

        config_parser = configparser.ConfigParser()
        config_parser.read(path_init_file)

        for section in config_parser.sections():
            for k, v in config_parser[section].items():
                # Select validator function and update dictionary
                validate_func = _validators[k]
                self.__setitem__(k, validate_func(v))


# Generate default config dictionary
config = GeoUtilsConfigDict()
config._set_defaults(path_init_file=_config_ini_file)



"""Test configuration."""

import logging
from typing import Any

import pytest


class LoggingWarningCollector(logging.Handler):
    """Helper class to collect logging warnings."""

    def __init__(self) -> None:
        super().__init__(level=logging.WARNING)
        self.records = []

    def emit(self, record: Any) -> None:
        self.records.append(record)


@pytest.fixture(autouse=True)  # type: ignore
def fail_on_logging_warnings(request: Any) -> Any:
    """Fixture used automatically in all tests to fail when a logging exceptions of WARNING or above is raised."""

    # The collector is required to avoid teardown, hookwrapper or plugin issues (we collect and fail later)
    collector = LoggingWarningCollector()
    root = logging.getLogger()
    root.addHandler(collector)

    # Run test
    yield

    root.removeHandler(collector)

    # Allow opt-out
    if request.node.get_closest_marker("allow_logging_warnings"):
        return

    # Categorize bad tests
    # IGNORED = ("rasterio",)   # If we want to add a list of "IGNORED" packages in the future
    bad = [
        r
        for r in collector.records
        if r.levelno >= logging.WARNING
        # and not r.name.startswith(IGNORED)
    ]

    # Fail on those exceptions and report exception level, name and message
    if bad:
        msgs = "\n".join(f"{r.levelname}:{r.name}:{r.getMessage()}" for r in bad)
        pytest.fail(
            "Logging warning/error detected:\n" + msgs,
            pytrace=False,
        )



from __future__ import annotations

import warnings
from cmath import isnan
from typing import Any

import numpy as np
import pytest
import rasterio as rio

import geoutils as gu
from geoutils import examples
from geoutils._typing import NDArrayNum

expected_stats = [
    "Mean",
    "Median",
    "Max",
    "Min",
    "Sum",
    "Sum of squares",
    "90th percentile",
    "IQR",
    "LE90",
    "NMAD",
    "RMSE",
    "Standard deviation",
]

expected_stats_count = [
    "Valid count",
    "Total count",
    "Percentage valid points",
]

expected_stats_mask = [
    "Valid inlier count",
    "Total inlier count",
    "Percentage inlier points",
    "Percentage valid inlier points",
]

stat_types = (int, float, np.integer, np.floating)


def compare_dict(dict1: dict, dict2: dict) -> None:  # type: ignore
    assert len(dict1.keys()) == len(dict1.keys())
    for key in dict1.keys():
        assert key in dict2
        if dict1[key] is not np.nan:
            assert dict2[key] == pytest.approx(dict1[key], abs=1e-10)
        else:
            assert dict2[key] is np.nan


class TestStats:
    landsat_b4_path = examples.get_path_test("everest_landsat_b4")
    landsat_rgb_path = examples.get_path_test("everest_landsat_rgb")
    aster_dem_path = examples.get_path_test("exploradores_aster_dem")

    @pytest.mark.parametrize("example", [landsat_b4_path, landsat_rgb_path, aster_dem_path])  # type: ignore
    def test_get_stats_raster(self, example: str) -> None:
        """
        Verify get_stats() method for a raster, especially output stats for different inputs
        parameters and some stats.
        """
        raster = gu.Raster(example)

        # Full stats
        stats = raster.get_stats()
        assert len(stats) == len(expected_stats + expected_stats_count)
        for name in expected_stats + expected_stats_count:
            assert name in stats
            assert isinstance(stats.get(name), stat_types)

        # With mask (inlier=True)
        inlier_mask = ~raster.get_mask()
        stats_masked = raster.get_stats(inlier_mask=inlier_mask)
        assert len(stats_masked) == len(expected_stats + expected_stats_count + expected_stats_mask)
        for name in expected_stats_mask:
            assert name in stats_masked
            stats_masked.pop(name)
        assert stats_masked == stats

        # Empty mask (=False)
        empty_mask = np.zeros_like(inlier_mask)
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning, message="Empty raster")
            stats_masked = raster.get_stats(inlier_mask=empty_mask)
        assert len(stats_masked) == len(expected_stats + expected_stats_count + expected_stats_mask)
        for name in expected_stats:
            assert np.isnan(stats_masked.get(name))

        assert stats_masked.get("Valid count") == stats.get("Valid count")
        assert stats_masked.get("Total count") == stats.get("Total count")
        assert stats_masked.get("Percentage valid points") == stats.get("Percentage valid points")

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning, message="Empty raster")
            stats_masked = raster.get_stats(inlier_mask=empty_mask, stats_name="mean")
        assert np.isnan(stats_masked)
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning, message="Empty raster")
            stats_masked = raster.get_stats(inlier_mask=empty_mask, stats_name="valid_count")
        assert stats_masked == stats.get("Valid count")
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning, message="Empty raster")
            stats_masked = raster.get_stats(inlier_mask=empty_mask, stats_name="Valid inlier count")
        assert stats_masked == 0
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning, message="Empty raster")
            stats_masked = raster.get_stats(inlier_mask=empty_mask, stats_name="validinliercount")
        assert stats_masked == 0

        # Empty DEM
        dem_empty = gu.Raster.from_array(
            np.random.randint(42, size=(0, 0), dtype="uint8"),
            transform=rio.transform.from_origin(10, 20, 1, 1),
            crs=4326,
        )
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning, message="Empty raster")
            stats_empty = dem_empty.get_stats()
        assert len(stats_empty) == len(expected_stats + expected_stats_count)
        for name in expected_stats:
            assert np.isnan(stats_empty.get(name))
        assert stats_empty.get("Valid count") == 0
        assert stats_empty.get("Total count") == 0
        assert isnan(stats_empty.get("Percentage valid points"))

        # Single stat
        for name in expected_stats + expected_stats_count:
            stat = raster.get_stats(stats_name=name)
            assert np.isfinite(stat)

        # Alias stat
        assert raster.get_stats(stats_name="Valid count") == raster.get_stats(stats_name="valid_count")

        # Callable
        def percentile_95(data: NDArrayNum) -> np.floating[Any]:
            if isinstance(data, np.ma.MaskedArray):
                data = data.compressed()
            return np.nanpercentile(data, 95)

        stat = raster.get_stats(stats_name=percentile_95)
        assert isinstance(stat, np.floating)

        # Selected stats and callable
        stats_name = ["mean", "max", "std", "validinliercount", "percentile_95"]
        stats = raster.get_stats(
            stats_name=["mean", "max", "std", "validinliercount", percentile_95], inlier_mask=inlier_mask
        )
        assert len(stats) == len(stats_name)
        for name in stats_name:
            assert name in stats
            assert not isnan(stats.get(name))

        # Non-existing stats
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="Statistic name 80 percentile is not recognized")
            stat = raster.get_stats(stats_name="80 percentile")
        assert isnan(stat)

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="Statistic name 42 is a not recognized string")
            stat = raster.get_stats(stats_name=42)
        assert stat is None

        # IQR (scipy) validation with numpy
        nan_arr = raster.get_nanarray()
        if nan_arr.ndim == 3:
            nan_arr = nan_arr[0, :, :]
        assert raster.get_stats(stats_name="iqr") == pytest.approx(
            np.nanpercentile(nan_arr, 75) - np.nanpercentile(nan_arr, 25)
        )

    @pytest.mark.parametrize("example", [landsat_b4_path, aster_dem_path])  # type: ignore
    def test_get_stats_raster_pointcloud(self, example: str) -> None:
        """
        Verify get_stats() method for a raster converted to pointcloud, especially output stats for different inputs
        parameters.
        """
        raster = gu.Raster(example)

        # Full stats
        stats = raster.to_pointcloud().get_stats()
        assert len(stats) == len(expected_stats + expected_stats_count)
        for name in expected_stats + expected_stats_count:
            assert name in stats
            assert isinstance(stats.get(name), stat_types)

        # Single stat
        for name in expected_stats + expected_stats_count:
            stat = raster.to_pointcloud().get_stats(stats_name=name)
            assert np.isfinite(stat)

        # Callable
        def percentile_95(data: NDArrayNum) -> np.floating[Any]:
            if isinstance(data, np.ma.MaskedArray):
                data = data.compressed()
            return np.nanpercentile(data, 95)

        # Selected stats and callable
        stats_name = ["mean", "max", "std", "percentile_95"]
        stats = raster.to_pointcloud().get_stats(stats_name=["mean", "max", "std", percentile_95])
        assert len(stats) == len(stats_name)
        for name in stats_name:
            assert name in stats
            assert stats.get(name) is not None

        # Non-existing stats
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="Statistic name 80 percentile is not recognized")
            stat = raster.get_stats(stats_name="80 percentile")
            assert isnan(stat)

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="Statistic name 42 is a not recognized string")
            stat = raster.get_stats(stats_name=42)
            assert stat is None

        # Empty mask (=False)
        inlier_mask = ~raster.get_mask()
        empty_mask = np.zeros_like(inlier_mask)
        raster.set_mask(~empty_mask)
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="Empty raster")
            stats_masked = raster.to_pointcloud().get_stats()
        assert len(stats_masked) == len(expected_stats + expected_stats_count)
        for name in expected_stats:
            assert np.isnan(stats_masked.get(name))
        assert stats_masked.get("Valid count") == 0
        assert stats_masked.get("Total count") == 0
        assert isnan(stats_masked.get("Percentage valid points"))

    def test_raster_get_stats_values(self) -> None:
        """
        Verify the output statistics values of a raster.
        """
        filename_rast = gu.examples.get_path("everest_landsat_b4")
        filename_vect = gu.examples.get_path("everest_rgi_outlines")
        rast = gu.Raster(filename_rast)
        vect = gu.Vector(filename_vect)
        inlier_mask = ~vect.create_mask(rast)

        # Verify raster stats
        res_stats = {
            "Mean": np.float64(144.04460496183205),
            "Median": np.float64(124.0),
            "Max": np.uint8(255),
            "Min": np.uint8(13),
            "Sum": np.uint64(75479373),
            "Sum of squares": np.uint64(44549637),
            "90th percentile": np.float64(255.0),
            "LE90": np.float64(218.0),
            "IQR": np.float64(164.0),
            "NMAD": np.float64(94.8864),
            "RMSE": np.float64(9.220541807365446),
            "Standard deviation": np.float64(79.44349437534403),
            "Valid count": 524000,
            "Total count": 524000,
            "Percentage valid points": np.float64(100.0),
        }
        compare_dict(res_stats, rast.get_stats())

        # Verify raster stats with a mask
        res_stats_mask = {
            "Mean": np.float64(110.49218069801574),
            "Median": np.float64(92.0),
            "Max": np.uint8(255),
            "Min": np.uint8(13),
            "Sum": np.uint64(26650493),
            "Sum of squares": np.uint64(24696991),
            "90th percentile": np.float64(225.0),
            "LE90": np.float64(223.0),
            "IQR": np.float64(83.0),
            "NMAD": np.float64(54.856199999999994),
            "RMSE": np.float64(10.118943490060417),
            "Standard deviation": np.float64(64.98157041836747),
            "Valid count": 524000,
            "Total count": 524000,
            "Percentage valid points": np.float64(100.0),
            "Valid inlier count": np.int64(241198),
            "Total inlier count": np.int64(241198),
            "Percentage inlier points": np.float64(46.03015267175572),
            "Percentage valid inlier points": np.float64(100.0),
        }
        compare_dict(res_stats_mask, rast.get_stats(inlier_mask=inlier_mask))

        # Verify cropped raster
        nrows, ncols = rast.shape
        rast_crop = rast.icrop((100, 100, ncols - 100, nrows - 100))
        res_stats_crop = {
            "Mean": np.float64(148.69901465201465),
            "Median": np.float64(133.0),
            "Max": np.uint8(255),
            "Min": np.uint8(14),
            "Sum": np.uint64(40594831),
            "Sum of squares": np.uint64(22875807),
            "90th percentile": np.float64(255.0),
            "LE90": np.float64(218.0),
            "IQR": np.float64(166.0),
            "NMAD": np.float64(105.26459999999999),
            "RMSE": np.float64(9.153915273540871),
            "Standard deviation": np.float64(79.32951386752386),
            "Valid count": 273000,
            "Total count": 273000,
            "Percentage valid points": np.float64(100.0),
        }
        compare_dict(res_stats_crop, rast_crop.get_stats())

        # Verify reprojected raster
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning, message="New nodata.*")
            rast_crop.set_nodata(255)  # Needs to be defined for reprojection
        rast_crop_proj = rast_crop.reproject(rast, resampling=rio.warp.Resampling.nearest)
        res_stats_crop_proj = {
            "Mean": np.float64(117.80631314205752),
            "Median": np.float64(107.0),
            "Max": np.uint8(254),
            "Min": np.uint8(14),
            "Sum": np.uint64(24919216),
            "Sum of squares": np.uint64(22814334),
            "90th percentile": np.float64(218.0),
            "LE90": np.float64(204.0),
            "IQR": np.float64(93.0),
            "NMAD": np.float64(66.717),
            "RMSE": np.float64(10.38534653788665),
            "Standard deviation": np.float64(62.319986152883956),
            "Valid count": 211527,
            "Total count": 524000,
            "Percentage valid points": np.float64(40.36774809160305),
        }
        compare_dict(res_stats_crop_proj, rast_crop_proj.get_stats())

        # Verify stats of a masked raster
        rast.set_mask(inlier_mask)
        stats_masked_rast = {
            "Mean": np.float64(172.66101371277432),
            "Median": np.float64(188.0),
            "Max": np.uint8(255),
            "Min": np.uint8(15),
            "Sum": np.uint64(48828880),
            "Sum of squares": np.uint64(19852646),
            "90th percentile": np.float64(255.0),
            "LE90": np.float64(209.0),
            "IQR": np.float64(156.0),
            "NMAD": np.float64(99.3342),
            "RMSE": np.float64(8.37853254688833),
            "Standard deviation": np.float64(79.45825061580675),
            "Valid count": 282802,
            "Total count": 524000,
            "Percentage valid points": np.float64(53.96984732824428),
        }
        compare_dict(stats_masked_rast, rast.get_stats())

        # Verify stats of a masked raster with the other part covered by the inler_mask (=> empty raster)
        stats_masked_rast_masked = {
            "Mean": np.nan,
            "Median": np.nan,
            "Max": np.nan,
            "Min": np.nan,
            "Sum": np.nan,
            "Sum of squares": np.nan,
            "90th percentile": np.nan,
            "LE90": np.nan,
            "IQR": np.nan,
            "NMAD": np.nan,
            "RMSE": np.nan,
            "Standard deviation": np.nan,
            "Valid count": 282802,
            "Total count": 524000,
            "Percentage valid points": np.float64(53.96984732824428),
            "Valid inlier count": np.int64(0),
            "Total inlier count": np.int64(241198),
            "Percentage inlier points": np.float64(0.0),
            "Percentage valid inlier points": np.float64(0.0),
        }
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning, message="Empty raster")
            compare_dict(stats_masked_rast_masked, rast.get_stats(inlier_mask=inlier_mask))

    def test_pointcloud_get_stats_values(self) -> None:
        """
        Verify the output statistics values of a pointcloud.
        """

        filename_rast = gu.examples.get_path("everest_landsat_b4")
        rast = gu.Raster(filename_rast)
        rast_pc = rast.to_pointcloud()

        # Verify pc stats
        rast_stats_pc = {
            "Mean": np.float64(144.04460496183205),
            "Median": np.float64(124.0),
            "Max": np.uint8(255),
            "Min": np.uint8(13),
            "Sum": np.uint64(75479373),
            "Sum of squares": np.uint64(44549637),
            "90th percentile": np.float64(255.0),
            "LE90": np.float64(218.0),
            "IQR": np.float64(164.0),
            "NMAD": np.float64(94.8864),
            "RMSE": np.float64(9.220541807365446),
            "Standard deviation": np.float64(79.44349437534403),
            "Valid count": 524000,
            "Total count": 524000,
            "Percentage valid points": np.float64(100.0),
        }
        compare_dict(rast_stats_pc, rast_pc.get_stats())

        # Verify cropped raster pc
        nrows, ncols = rast.shape
        rast_crop = rast.icrop((100, 100, ncols - 100, nrows - 100))
        rast_crop_pc = rast_crop.to_pointcloud()

        rast_stats_crop_pc = {
            "Mean": np.float64(148.69901465201465),
            "Median": np.float64(133.0),
            "Max": np.uint8(255),
            "Min": np.uint8(14),
            "Sum": np.uint64(40594831),
            "Sum of squares": np.uint64(22875807),
            "90th percentile": np.float64(255.0),
            "LE90": np.float64(218.0),
            "IQR": np.float64(166.0),
            "NMAD": np.float64(105.26459999999999),
            "RMSE": np.float64(9.153915273540871),
            "Standard deviation": np.float64(79.32951386752386),
            "Valid count": 273000,
            "Total count": 273000,
            "Percentage valid points": np.float64(100.0),
        }
        compare_dict(rast_stats_crop_pc, rast_crop_pc.get_stats())

        # Verify reprojected raster pc
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning, message="New nodata.*")
            rast_crop.set_nodata(255)  # Needs to be defined for reprojection
        rast_crop_proj = rast_crop.reproject(rast, resampling=rio.warp.Resampling.nearest)
        for data in [rast_crop_proj, rast_crop, rast]:
            print(
                "# data shape:", data.shape, data.shape[0] * rast_crop.shape[1], "->", len(data.to_pointcloud()["b1"])
            )
            print("    ", data.to_pointcloud()["b1"].min(), "to", data.to_pointcloud()["b1"].max())
        rast_crop_proj_pc = rast_crop_proj.to_pointcloud()

        rast_stats_crop_proj_pc = {
            "Mean": np.float64(117.80631314205752),
            "Median": np.float64(107.0),
            "Max": np.uint8(254),
            "Min": np.uint8(14),
            "Sum": np.uint64(24919216),
            "Sum of squares": np.uint64(22814334),
            "90th percentile": np.float64(218.0),
            "LE90": np.float64(204.0),
            "IQR": np.float64(93.0),
            "NMAD": np.float64(66.717),
            "RMSE": np.float64(10.38534653788665),
            "Standard deviation": np.float64(62.319986152883956),
            "Valid count": 211527,
            "Total count": 211527,
            "Percentage valid points": np.float64(100.0),
        }
        compare_dict(rast_stats_crop_proj_pc, rast_crop_proj_pc.get_stats())



"""Test distance functions at the interface of raster and vectors."""

from __future__ import annotations

import os
import tempfile
import warnings

import numpy as np
import pytest
import rasterio as rio

import geoutils as gu
from geoutils._typing import NDArrayNum


def run_gdal_proximity(
    input_raster: gu.Raster, target_values: list[float] | None, distunits: str = "GEO"
) -> NDArrayNum:
    """Run GDAL's ComputeProximity and return the read numpy array."""
    # Rasterio strongly recommends against importing gdal along rio, so this is done here instead.
    from osgeo import gdal, gdalconst

    gdal.UseExceptions()

    # Initiate empty GDAL raster for proximity output
    drv = gdal.GetDriverByName("MEM")
    proxy_ds = drv.Create("", input_raster.shape[1], input_raster.shape[0], 1, gdal.GetDataTypeByName("Float32"))
    proxy_ds.GetRasterBand(1).SetNoDataValue(-9999)

    # Save input in temporary file to read with GDAL
    # (avoids the nightmare of setting nodata, transform, crs in GDAL format...)
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = os.path.join(temp_dir, "input.tif")
        input_raster.to_file(temp_path)
        ds_raster_in = gdal.Open(temp_path, gdalconst.GA_ReadOnly)

        # Define GDAL options
        proximity_options = ["DISTUNITS=" + distunits]
        if target_values is not None:
            proximity_options.insert(0, "VALUES=" + ",".join([str(tgt) for tgt in target_values]))

        # Compute proximity
        gdal.ComputeProximity(ds_raster_in.GetRasterBand(1), proxy_ds.GetRasterBand(1), proximity_options)
        # Save array
        proxy_array = proxy_ds.GetRasterBand(1).ReadAsArray().astype("float32")
        proxy_array[proxy_array == -9999] = np.nan

        # Close GDAL datasets
        proxy_ds = None
        ds_raster_in = None

    return proxy_array


class TestDistance:

    landsat_b4_path = gu.examples.get_path_test("everest_landsat_b4")
    landsat_b4_crop_path = gu.examples.get_path_test("everest_landsat_b4_cropped")
    everest_outlines_path = gu.examples.get_path_test("everest_rgi_outlines")
    aster_dem_path = gu.examples.get_path_test("exploradores_aster_dem")

    def test_proximity_vector(self) -> None:
        """
        The core functionality is already tested against GDAL in test_raster: just verify the vector-specific behaviour.
        #TODO: add an artificial test as well (mirroring TODO in test_raster)
        """

        vector = gu.Vector(self.everest_outlines_path)

        # -- Test 1: with a Raster provided --
        raster1 = gu.Raster(self.landsat_b4_crop_path)
        prox1 = vector.proximity(raster=raster1)

        # The proximity should have the same extent, resolution and CRS
        assert raster1.georeferenced_grid_equal(prox1)

        # With the base geometry
        vector.proximity(raster=raster1, geometry_type="geometry")

        # With another geometry option
        vector.proximity(raster=raster1, geometry_type="centroid")

        # With only inside proximity
        vector.proximity(raster=raster1, in_or_out="in")

        # -- Test 2: with no Raster provided, just grid size --

        # Default grid size
        vector.proximity()

        # With specific grid size
        vector.proximity(size=(100, 100))

        # Test all options, with both an artificial Raster (that has all target values) and a real Raster

    @pytest.mark.parametrize("distunits", ["GEO", "PIXEL"])  # type: ignore
    # 0 and 1,2,3 are especially useful for the artificial Raster, and 112 for the real Raster
    @pytest.mark.parametrize("target_values", [[1, 2, 3], [0], [112], None])  # type: ignore
    @pytest.mark.parametrize(
        "raster",
        [
            gu.Raster(landsat_b4_path),
            gu.Raster.from_array(
                np.arange(25, dtype="int32").reshape(5, 5), transform=rio.transform.from_origin(0, 5, 1, 1), crs=4326
            ),
        ],
    )  # type: ignore
    def test_proximity_raster_against_gdal(
        self, distunits: str, target_values: list[float] | None, raster: gu.Raster
    ) -> None:
        """Test that proximity matches the results of GDAL for any parameter."""

        # TODO: When adding new rasters for tests, specify warning only for Landsat
        warnings.filterwarnings("ignore", message="Setting default nodata -99999 to mask non-finite values *")

        # We generate proximity with GDAL and GeoUtils
        gdal_proximity = run_gdal_proximity(raster, target_values=target_values, distunits=distunits)
        # We translate distunits GDAL option into its GeoUtils equivalent
        if distunits == "GEO":
            distance_unit = "georeferenced"
        else:
            distance_unit = "pixel"
        geoutils_proximity = (
            raster.proximity(distance_unit=distance_unit, target_values=target_values)
            .data.data.squeeze()
            .astype("float32")
        )

        # The results should be the same in all cases
        try:
            # In some cases, the proximity differs slightly (generally <1%) for complex settings
            # (Landsat Raster with target of 112)
            # It looks like GDAL might not have the right value,
            # so this particular case is treated differently in tests
            if target_values is not None and target_values[0] == 112 and raster.filename is not None:
                # Get index and number of not almost equal point (tolerance of 10-4)
                ind_not_almost_equal = np.abs(gdal_proximity - geoutils_proximity) > 1e-04
                nb_not_almost_equal = np.count_nonzero(ind_not_almost_equal)
                # Check that this is a minority of points (less than 0.5%)
                assert nb_not_almost_equal < 0.005 * raster.width * raster.height

                # Replace these exceptions by zero in both
                gdal_proximity[ind_not_almost_equal] = 0.0
                geoutils_proximity[ind_not_almost_equal] = 0.0
                # Check that all the rest is almost equal
                assert np.allclose(gdal_proximity, geoutils_proximity, atol=1e-04, equal_nan=True)

            # Otherwise, results are exactly equal
            else:
                assert np.array_equal(gdal_proximity, geoutils_proximity, equal_nan=True)

        # For debugging
        except Exception as exception:
            import matplotlib.pyplot as plt

            # Plotting the xdem and GDAL attributes for comparison (plotting "diff" can also help debug)
            plt.subplot(121)
            plt.imshow(gdal_proximity)
            # plt.imshow(np.abs(gdal_proximity - geoutils_proximity)>0.1)
            plt.colorbar()
            plt.subplot(122)
            plt.imshow(geoutils_proximity)
            # plt.imshow(raster.data.data == 112)
            plt.colorbar()
            plt.show()

            # ind_not_equal = np.abs(gdal_proximity - geoutils_proximity)>0.1
            # print(gdal_proximity[ind_not_equal])
            # print(geoutils_proximity[ind_not_equal])

            raise exception

    def test_proximity_raster_parameters(self) -> None:
        """
        Test that new (different to GDAL's) proximity parameters run.
        No need to test the results specifically, as those rely entirely on the previous test with GDAL,
        and tests in rasterize and shapely.
        #TODO: Maybe add one test with an artificial vector to check it works as intended
        """

        # -- Test 1: with self's Raster alone --
        raster1 = gu.Raster(self.landsat_b4_path)
        prox1 = raster1.proximity()

        # The raster should have the same extent, resolution and CRS
        assert raster1.georeferenced_grid_equal(prox1)

        # It should change with target values specified
        prox2 = raster1.proximity(target_values=[255])
        assert not np.array_equal(prox1.data, prox2.data)

        # -- Test 2: with a vector provided --
        vector = gu.Vector(self.everest_outlines_path)

        # With default options (boundary geometry)
        raster1.proximity(vector=vector)

        # With the base geometry
        raster1.proximity(vector=vector, geometry_type="geometry")

        # With another geometry option
        raster1.proximity(vector=vector, geometry_type="centroid")

        # With only inside proximity
        raster1.proximity(vector=vector, in_or_out="in")

        # Paths to example data

    # Mask without nodata
    mask_landsat_b4 = gu.Raster(landsat_b4_path) > 125
    # Mask with nodata
    mask_aster_dem = gu.Raster(aster_dem_path) > 2000
    # Mask from an outline
    mask_everest = gu.Vector(everest_outlines_path).create_mask(gu.Raster(landsat_b4_path))

    @pytest.mark.parametrize("mask", [mask_landsat_b4, mask_aster_dem, mask_everest])  # type: ignore
    def test_proximity_mask(self, mask: gu.Raster) -> None:
        mask_orig = mask.copy()
        # Run default
        rast = mask.proximity()
        # Check the dtype of the original mask was properly reconverted
        assert mask.data.dtype == bool
        # Check the original mask was not modified during reprojection
        assert mask_orig.raster_equal(mask)

        # Check that output is cast back into a raster
        assert isinstance(rast, gu.Raster)



"""
Open/save a point cloud
=======================

This example demonstrates the instantiation of a point cloud through :class:`geoutils.PointCloud` and saving with :func:`~geoutils.Vector.save`.
"""

import geoutils as gu

# %%
# We open an example vector.
filename_pc = gu.examples.get_path("coromandel_lidar")
pc = gu.PointCloud(filename_pc, data_column="Z")
pc

# %%
# A point cloud is a subclass of :class:`~geoutils.Vector`, with a main attribute :attr:`~geoutils.PointCloud.data_column` pointing to the main data column
# of the point cloud.
# All other attributes are :ref:`inherited from Shapely and GeoPandas<vector-from-geopandas>`. See also :ref:`vector-class`.

# %%
#
# .. note::
#        A point cloud can also be instantiated with a :class:`geopandas.GeoDataFrame`, see :ref:`sphx_glr_io_examples_import_export_import_vector.py`.
#
# We can print more info on the point cloud.
pc.info()

# %%
# Let's plot the point cloud main column
pc.plot(cbar_title="Elevation (m)")

# %%
# Finally, a point cloud is saved using :func:`~geoutils.Vector.save`.

pc.to_file("mypc.gpkg")



"""
Crop a raster
=============

This example demonstrates the cropping of a raster using :func:`geoutils.Raster.crop`.
"""

# %%
# We open a raster and vector, and subset the latter.

# sphinx_gallery_thumbnail_number = 2
import geoutils as gu

filename_rast = gu.examples.get_path("everest_landsat_b4")
filename_vect = gu.examples.get_path("everest_rgi_outlines")
rast = gu.Raster(filename_rast)
vect = gu.Vector(filename_vect)
vect = vect[vect["RGIId"] == "RGI60-15.10055"]

# %%
# The first raster has larger extent and higher resolution than the vector.
rast.info()
print(vect.bounds)

# %%
# Let's plot the raster and vector.
rast.plot(cmap="Purples")
vect.plot(ref_crs=rast, fc="none", ec="k", lw=2)

# %%
# **First option:** using the vector as a reference to match, we reproject the raster. We simply have to pass the :class:`~geoutils.Vector`
# as single argument to :func:`~geoutils.Raster.crop`. See :ref:`core-match-ref` for more details.

rast = rast.crop(vect)

# %%
# Now the bounds should be the same as that of the vector (within the size of a pixel as the grid was not warped).
#
# .. note::
#      By default, :func:`~geoutils.Raster.crop` is done in-place, replacing ``rast``. This behaviour can be modified by passing ``inplace=False``.
#

rast.plot(ax="new", cmap="Purples")
vect.plot(ref_crs=rast, fc="none", ec="k", lw=2)

# %%
# **Second option:** we can pass other ``crop_geom`` argument to :func:`~geoutils.Raster.crop`, including another :class:`~geoutils.Raster` or a
# simple :class:`tuple` of bounds. For instance, we can re-crop the raster to be smaller than the vector.

rast = rast.crop((rast.bounds.left + 1000, rast.bounds.bottom, rast.bounds.right, rast.bounds.top - 500))

rast.plot(ax="new", cmap="Purples")
vect.plot(ref_crs=rast, fc="none", ec="k", lw=2)



"""
Tests for multiprocessing functions
"""

import os
import warnings
from multiprocessing import cpu_count
from typing import Any

import numpy as np
import pytest
import rasterio as rio
import scipy
from numpy import floating

import geoutils as gu
from geoutils import Raster, examples
from geoutils.raster import RasterType
from geoutils.raster.distributed_computing import (
    AbstractCluster,
    ClusterGenerator,
    MultiprocConfig,
    map_multiproc_collect,
    map_overlap_multiproc_save,
)
from geoutils.raster.distributed_computing.multiproc import (
    _apply_func_block,
    _load_raster_tile,
    _remove_tile_padding,
)


# Define a simple function where overlap is needed
def _custom_func_overlap(raster: RasterType, size: int) -> RasterType:
    new_data = scipy.ndimage.maximum_filter(raster.data, size)
    if raster.nodata is not None:
        new_data = np.ma.masked_equal(new_data, raster.nodata)
    raster.data = new_data
    return raster


# Define a simple function with some args
def _custom_func(raster: RasterType, addition: float, factor: float) -> RasterType:
    return (raster + addition) * factor


# Define a simple function which do not return a Raster
def _custom_func_stats(raster: RasterType) -> dict[str, floating[Any]]:
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning, message="Empty raster")
        return raster.get_stats(stats_name=["mean", "valid_count"])


# Define a simple function which return a Mask
def _custom_func_mask(raster: RasterType) -> gu.Raster:
    mask_array = raster.get_mask()
    return gu.Raster.from_array(mask_array, raster.transform, raster.crs)


class TestMultiproc:
    aster_dem_path = examples.get_path_test("exploradores_aster_dem")
    landsat_rgb_path = examples.get_path_test("everest_landsat_rgb")

    num_workers = min(2, cpu_count())  # Safer limit for CI
    cluster = ClusterGenerator("test", nb_workers=num_workers)

    @pytest.mark.parametrize("example", [aster_dem_path, landsat_rgb_path])  # type: ignore
    def test_load_raster_tile(self, example: str) -> None:
        """
        Test loading a specific tile (spatial subset) from the raster.
        """
        raster = Raster(example)
        # Define a tile (bounding box) to load
        tile = np.array([50, 125, 100, 200])  # [rowmin, rowmax, colmin, colmax]

        # Load the tile and verify dimensions
        raster_tile = _load_raster_tile(raster, tile)
        assert np.array_equal(raster_tile.data, raster.data[..., tile[0] : tile[1], tile[2] : tile[3]])

    @pytest.mark.parametrize("example", [aster_dem_path, landsat_rgb_path])  # type: ignore
    @pytest.mark.parametrize("padding", [0, 1, 10])  # type: ignore
    def test_remove_tile_padding(self, example: str, padding: int) -> None:
        """
        Test removing padding from a raster tile after processing.
        """
        raster = Raster(example)
        # Extract a tile with padding
        tile = np.array([0, 100, 50, 150])
        tile_pad = tile + np.array([-1, 1, -1, 1]) * padding

        raster_tile = _load_raster_tile(raster, tile)
        raster_tile_with_padding = _load_raster_tile(raster, tile_pad)

        # Remove padding and ensure it's back to the original size
        _remove_tile_padding((raster.height, raster.width), raster_tile_with_padding, tile, padding)
        assert raster_tile_with_padding.raster_equal(raster_tile)

    @pytest.mark.parametrize("example", [aster_dem_path, landsat_rgb_path])  # type: ignore
    @pytest.mark.parametrize("padding", [0, 1, 3])  # type: ignore
    def test_apply_func_block(self, example: str, padding: int) -> None:
        """
        Test applying a function to a raster tile and handling padding removal.
        """
        raster = Raster(example)
        tile = np.array([10, 20, 10, 20])  # [rowmin, rowmax, colmin, colmax]
        size = 2

        # Apply map_block
        result_tile, _ = _apply_func_block(_custom_func_overlap, raster, tile, padding, size)

        raster = _custom_func_overlap(raster, size)
        # If padding >=1, The result should be the equal to the original tile filtered
        original_tile_filtered = _load_raster_tile(raster, tile)
        if padding >= size - 1:
            assert result_tile.raster_equal(original_tile_filtered)
        else:
            assert not result_tile.raster_equal(original_tile_filtered)

    @pytest.mark.parametrize("example", [aster_dem_path, landsat_rgb_path])  # type: ignore
    @pytest.mark.parametrize("tile_size", [100, 200])  # type: ignore
    @pytest.mark.parametrize("cluster", [None, cluster])  # type: ignore
    def test_map_overlap_multiproc_save(self, example: str, tile_size: int, cluster: None | AbstractCluster) -> None:
        """
        Test the multiprocessing map function with a simple operation returning a raster.
        """
        raster = Raster(example)
        output_file = "output.tif"
        depth = 10
        config = MultiprocConfig(tile_size, output_file, cluster=cluster)

        addition = 5
        factor = 0.5
        # Apply the multiproc map function

        output_raster = map_overlap_multiproc_save(_custom_func, raster, config, addition, factor, depth=depth)

        # Ensure raster has not been loading during process
        assert not raster.is_loaded

        # Ensure the output file is created and valid
        assert os.path.exists(output_file)

        # Compare with the operation on full raster
        new_raster = _custom_func(raster, addition, factor)
        assert output_raster.raster_equal(new_raster)

        # Assert raster has been properly saved in output_file
        output_raster_saved = Raster(output_file)
        assert output_raster_saved.raster_equal(output_raster)

        # Remove output file
        os.remove(output_file)

        # With a tempfile :
        config = MultiprocConfig(tile_size)
        output_raster = map_overlap_multiproc_save(_custom_func, raster, config, addition, factor, depth=depth)
        output_raster_saved = Raster(config.outfile)
        assert output_raster_saved.raster_equal(output_raster)

        if raster.count == 1:
            # With a wrapper returning a Mask
            output_mask = map_overlap_multiproc_save(_custom_func_mask, raster, config, depth=depth)
            assert np.array_equal(raster.get_mask(), output_mask.data)

    @pytest.mark.parametrize("example", [aster_dem_path, landsat_rgb_path])  # type: ignore
    @pytest.mark.parametrize("tile_size", [10, 20])  # type: ignore
    @pytest.mark.parametrize("cluster", [None, cluster])  # type: ignore
    @pytest.mark.parametrize("return_tile", [False, True])  # type: ignore
    def test_map_multiproc_collect(
        self, example: str, tile_size: int, cluster: None | AbstractCluster, return_tile: bool
    ) -> None:
        """
        Test the multiprocessing map function with a simple operation returning not a raster.
        """
        raster = Raster(example)
        config = MultiprocConfig(tile_size, cluster=cluster)

        # Apply the multiproc map function
        results = map_multiproc_collect(_custom_func_stats, raster, config, return_tile=return_tile)  # type: ignore
        if return_tile:
            list_stats = [result[0] for result in results]
            list_tiles = [result[1] for result in results]
            assert np.array_equal(list_tiles[0], np.array([0, tile_size, 0, tile_size]))
        else:
            list_stats = results

        # Ensure raster has not been loading during process
        assert not raster.is_loaded

        # Compare tiled_stats with the stats on full raster
        total_stats = _custom_func_stats(raster)

        tiled_count = np.nansum([stats["valid_count"] for stats in list_stats])
        tiled_mean = np.nansum([stats["mean"] * stats["valid_count"] for stats in list_stats]) / tiled_count
        assert abs(total_stats["mean"] - tiled_mean) < tiled_mean * 1e-5
        assert total_stats["valid_count"] == tiled_count

    @pytest.mark.skip()  # type: ignore
    @pytest.mark.parametrize("example", [aster_dem_path])  # type: ignore
    @pytest.mark.parametrize("tile_size", [10, 20])  # type: ignore
    @pytest.mark.parametrize("cluster", [None, cluster])  # type: ignore
    def test_multiproc_reproject(self, example: str, tile_size: int, cluster: None | AbstractCluster) -> None:
        """Test for multiproc_reproject"""

        r = gu.Raster(example)
        config = MultiprocConfig(tile_size, cluster=cluster)

        # specific for the landsat test case, default nodata 255 cannot be used (see above), so use 0
        if r.nodata is None:
            r.set_nodata(0)

        # - Test reprojection with bounds and resolution -
        dst_bounds = rio.coords.BoundingBox(
            left=r.bounds.left, bottom=r.bounds.bottom + r.res[0], right=r.bounds.right - 2 * r.res[1], top=r.bounds.top
        )
        res_tuple = (r.res[0] * 0.5, r.res[1] * 3)

        # Multiprocessing reprojection
        r_multi = r.reproject(bounds=dst_bounds, res=res_tuple, multiproc_config=config)

        # Assert that the raster has not been loaded during reprojection
        assert not r.is_loaded

        # Single-process reprojection
        r_single = r.reproject(bounds=dst_bounds, res=res_tuple)

        # Assert the results are the same
        assert r_single.raster_equal(r_multi)

        # - Test reprojection with CRS change -
        for out_crs in [rio.crs.CRS.from_epsg(4326)]:

            # Single-process reprojection
            r_single = r.reproject(crs=out_crs)

            # Multiprocessing reprojection
            r_multi = r.reproject(crs=out_crs, multiproc_config=config)

            # Assert the results are the same
            assert r_single.raster_equal(r_multi)

        # Check that reprojection works for several bands in multiproc as well
        for n in [2, 3, 4]:
            img1 = Raster.from_array(
                np.ones((n, 500, 500), dtype="uint8"),
                transform=rio.transform.from_origin(0, 500, 1, 1),
                crs=4326,
                nodata=0,
            )
            img2 = Raster.from_array(
                np.ones((n, 500, 500), dtype="uint8"),
                transform=rio.transform.from_origin(50, 500, 1, 1),
                crs=4326,
                nodata=0,
            )

            out_img_single = img2.reproject(img1)
            out_img_multi = img2.reproject(ref=img1, multiproc_config=config)

            assert out_img_multi.count == n
            assert out_img_multi.shape == (500, 500)
            assert out_img_single.raster_equal(out_img_multi)

    @pytest.mark.skip()  # type: ignore
    def test_map_overlap_multiproc_save_bigTiff(self) -> None:
        """
        Test the multiprocessing map function with a simple operation returning a raster > 4go (BigTIFF format)
        """



# Copyright (c) 2025 GeoUtils developers
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for zonal statistics."""

from __future__ import annotations

import warnings
from collections.abc import Callable
from functools import partial
from typing import Any

import numpy as np
from scipy.stats import iqr
from scipy.stats.mstats import mquantiles

from geoutils import profiler
from geoutils._typing import NDArrayNum
from geoutils.stats.estimators import linear_error, nmad, rmse, sum_square

_STATS_ALIASES = {
    "mean": "Mean",
    "median": "Median",
    "max": "Max",
    "maximum": "Max",
    "min": "Min",
    "minimum": "Min",
    "sum": "Sum",
    "sumofsquares": "Sum of squares",
    "sum2": "Sum of squares",
    "90thpercentile": "90th percentile",
    "90percentile": "90th percentile",
    "iqr": "IQR",
    "le90": "LE90",
    "nmad": "NMAD",
    "rmse": "RMSE",
    "rms": "RMSE",
    "std": "Standard deviation",
    "standarddeviation": "Standard deviation",
    "standard_deviation": "Standard deviation",
    "validcount": "Valid count",
    "valid_count": "Valid count",
    "totalcount": "Total count",
    "total_count": "Total count",
    "percentagevalidpoints": "Percentage valid points",
    "percentage_valid_points": "Percentage valid points",
}  # type: ignore

STATS_LIST = [
    "Mean",
    "Median",
    "Max",
    "Min",
    "Sum",
    "Sum of squares",
    "90th percentile",
    "IQR",
    "LE90",
    "NMAD",
    "RMSE",
    "Standard deviation",
    "Valid count",
    "Total count",
    "Percentage valid points",
]

STATS_LIST_MASK = [
    "Valid inlier count",
    "Total inlier count",
    "Percentage inlier points",
    "Percentage valid inlier points",
]

_ALIAS_STATS_LIST_MASK = {
    "validinliercount": "Valid inlier count",
    "valid_inlier_count": "Valid inlier count",
    "totalinliercount": "Total inlier count",
    "total_inlier_count": "Total inlier count",
    "percentagevalidinlierpoints": "Percentage valid inlier points",
    "percentage_valid_inlier_points": "Percentage valid inlier points",
    "percentageinlierpoints": "Percentage inlier points",
    "percentage_inlier_points": "Percentage inlier points",
}


@profiler.profile("geoutils.stats.stats._statistics", memprof=True)  # type: ignore
def _statistics(
    data: NDArrayNum,
    stats_name: list[str | Callable[[NDArrayNum], np.floating[Any]]] | None = None,
    counts: tuple[int, int] | None = None,
) -> dict[str, float]:
    """
    Calculate common statistics for an N-D array :

    - Mean: arithmetic mean of the data, ignoring masked values.
    - Median: middle value when the valid data points are sorted in increasing order, ignoring masked values.
    - Max: maximum value among the data, ignoring masked values.
    - Min: minimum value among the data, ignoring masked values.
    - Sum: sum of all data, ignoring masked values.
    - Sum of squares: sum of the squares of all data, ignoring masked values.
    - 90th percentile: point below which 90% of the data falls, ignoring masked values.
    - IQR (Interquartile Range): difference between the 75th and 25th percentile of a dataset, ignoring masked values.
    - LE90 (Linear Error with 90% confidence): difference between the 95th and 5th percentiles of a dataset, \
    representing the range within which 90% of the data points lie. Ignore masked values.
    - NMAD (Normalized Median Absolute Deviation): robust measure of variability in the data, less sensitive to \
    outliers compared to standard deviation. Ignore masked values.
    - RMSE (Root Mean Square Error): commonly used to express the magnitude of errors or variability and can give \
    insight into the spread of the data. Only relevant when the raster represents a difference of two objects. \
    Ignore masked values.
    - Std (Standard deviation): measures the spread or dispersion of the data around the mean, ignoring masked values.
    - Valid count: number of finite data points in the array. It counts the non-masked elements.
    - Total count: total size of the raster.
    - Percentage valid points: ratio between Valid count and Total count.

    For all statistics up to and including "Std", NumPy Masked functions are used (directly or in the calculation)
    in case of a masked array, NumPy module otherwise.

    "Valid count" represents all non zero and not masked pixels in the input data (final_count_nonzero), previously
    calculated in case of a Raster.get_stats() called with an inlier_mask, before the mask application. NumPy Masked
    functions is used is this case or if the input was already a masked array.
    Percentage valid points is calculated accordingly.

    If an inlier mask is passed:
    - Total inlier count: number of data points in the inlier mask.
    - Valid inlier count: number of unmasked data points in the array after applying the inlier mask.
    - Percentage inlier points: ratio between Valid inlier count and Valid count. Useful for classification statistics.
    - Percentage valid inlier points: ratio between Valid inlier count and Total inlier count.

    They are all computed based on the previously stated final_count_nonzero.

    Callable functions are supported as well.

    :param data: Array on which to compute statistics.
    :param stats_name: list of names of the statistics to retrieve. If None, all statistics are returned.
            Accepted names include:
            `mean`, `median`, `max`, `min`, `sum`, `sum of squares`, `90th percentile`, `iqr`, `LE90`, `nmad`, `rmse`,
            `std`, `valid count`, `total count`, `percentage valid points` and if an inlier mask is passed :
            `valid inlier count`, `total inlier count`, `percentage inlier point`, `percentage valid inlier points`.
            Custom callables can also be provided.
    :param counts: Tuple with number of finite data points in array and number of valid points in inlier_mask.

    :returns: A dictionary containing the calculated statistics for the selected band.
    """

    if np.ma.isMaskedArray(data):

        # Count non zero and not masked pixels in the input data
        final_count_nonzero = np.count_nonzero(~np.ma.getmaskarray(data))

        # Compute valid count from non zero and not masked pixels in the input data
        # beforehand saved in counts[0] in case of a inler_mask parameter in get_stats()
        valid_count = final_count_nonzero if counts is None else counts[0]

        stats_dict = {
            "Mean": np.ma.mean,
            "Median": np.ma.median,
            "Max": np.ma.max,
            "Min": np.ma.min,
            "Sum": np.ma.sum,
            "Sum of squares": sum_square,
            "90th percentile": partial(lambda x: mquantiles(x, prob=0.9, alphap=1, betap=1)[0]),
            "LE90": partial(linear_error, interval=90),
            "IQR": partial(iqr, nan_policy="omit"),  # ignore masked value (nan),
            "NMAD": nmad,
            "RMSE": rmse,
            "Standard deviation": np.ma.std,
        }  # type: ignore

    else:
        # Count non zero pixels in the input data
        final_count_nonzero = np.count_nonzero(np.isfinite(data))

        # Compute valid count from non zero and not masked pixels in the input data
        # beforehand saved in counts[0] in case of a inler_mask parameter in get_stats()
        valid_count = final_count_nonzero if counts is None else counts[0]

        stats_dict = {
            "Mean": np.nanmean,
            "Median": np.nanmedian,
            "Max": np.nanmax,
            "Min": np.nanmin,
            "Sum": np.nansum,
            "Sum of squares": sum_square,
            "90th percentile": partial(np.nanpercentile, q=90),
            "LE90": partial(linear_error, interval=90),
            "IQR": partial(iqr, nan_policy="omit"),  # ignore masked value (nan),
            "NMAD": nmad,
            "RMSE": rmse,
            "Standard deviation": np.nanstd,
        }  # type: ignore

    # Pixels counts
    stats_dict.update(
        {
            "Valid count": valid_count,
            "Total count": data.size,
            "Percentage valid points": (valid_count / data.size) * 100 if data.size else np.nan,
        }
    )

    # If there are no valid data points, set all statistics to NaN
    if final_count_nonzero == 0:
        warnings.warn("Empty raster, returns Nan for all stats", category=UserWarning)
        if stats_name is None:
            stat_data_valid = STATS_LIST  # type: ignore
        else:
            stat_data_valid = stats_name  # type: ignore
        res_dict = {
            stat_name: (
                stats_dict[stat_name] if (stat_name in STATS_LIST and not callable(stats_dict[stat_name])) else np.nan
            )
            for stat_name in stat_data_valid
        }  # type: ignore

        if stats_name is not None:
            stat_with_alias = set(stats_name).intersection(list(_STATS_ALIASES.keys()))  # type: ignore
            if stat_with_alias:
                for stat_name in stat_with_alias:
                    alias = _STATS_ALIASES[stat_name]  # type: ignore
                    if callable(stats_dict[alias]):
                        res_dict[stat_name] = np.nan  # type: ignore
                    else:
                        res_dict[stat_name] = stats_dict[alias]  # type: ignore

    else:
        if stats_name is None:
            res_dict = stats_dict  # type: ignore
            if stats_name is None:
                for key in stats_dict.keys():
                    if callable(stats_dict[key]):
                        res_dict[key] = stats_dict[key](data)  # type: ignore
        else:
            res_dict = {}  # type: ignore
            for stat_name in stats_name:
                # Compute stat if in stats_dict keys
                if isinstance(stat_name, str) and stat_name in stats_dict.keys():
                    if callable(stats_dict[stat_name]):
                        res_dict[stat_name] = stats_dict[stat_name](data)  # type: ignore
                    else:
                        res_dict[stat_name] = stats_dict[stat_name]  # type: ignore
                # Compute stat if in _STATS_ALIASES keys
                elif isinstance(stat_name, str) and stat_name in _STATS_ALIASES.keys():
                    if callable(stats_dict[_STATS_ALIASES[stat_name]]):
                        res_dict[stat_name] = stats_dict[_STATS_ALIASES[stat_name]](data)  # type: ignore
                    else:
                        res_dict[stat_name] = stats_dict[_STATS_ALIASES[stat_name]]  # type: ignore
                # Compute stat if callable
                elif callable(stat_name):
                    res_dict[stat_name.__name__] = stat_name(data)  # type: ignore

                else:
                    # if none of the above conditions are met and if stats_name is not about the inlier mask
                    if stat_name not in STATS_LIST_MASK and stat_name not in _ALIAS_STATS_LIST_MASK:
                        warnings.warn("Statistic name " + stat_name + " is not recognized", category=UserWarning)
                        res_dict[stat_name] = np.float32(np.nan)  # type: ignore

    # If inlier mask parameter given before in get_stats() and if one of these stats is wanted
    if counts is not None and (
        stats_name is None
        or list(
            set(STATS_LIST_MASK).intersection(stats_name) or list(set(_ALIAS_STATS_LIST_MASK).intersection(stats_name))
        )
    ):
        dict_c = {
            "Valid inlier count": final_count_nonzero,
            "Total inlier count": counts[1],
            "Percentage inlier points": (final_count_nonzero / counts[0]) * 100,
            "Percentage valid inlier points": (final_count_nonzero / counts[1]) * 100 if counts[1] != 0 else 0,
        }

        if stats_name is None:
            # Add all stats
            res_dict.update(dict_c)
        else:
            # Add all stats if the name is on the list
            res_dict.update({k: dict_c[k] for k in list(set(STATS_LIST_MASK).intersection(stats_name))})
            # and the stats from alias stats
            res_dict.update(
                {
                    k: dict_c[_ALIAS_STATS_LIST_MASK[k]]
                    for k in list(set(_ALIAS_STATS_LIST_MASK).intersection(stats_name))
                }
            )
    return res_dict  # type: ignore



from __future__ import annotations

from collections.abc import Callable
from importlib.util import find_spec

import numpy as np
import pytest
import scipy

import geoutils as gu
from geoutils import Raster
from geoutils._typing import NDArrayNum
from geoutils.raster import get_array_and_mask


class TestGaussianFilter:
    """Tests for the Gaussian filter applied to raster data."""

    landsat_data = gu.Raster(gu.examples.get_path("everest_landsat_b4")).astype(np.float32)

    def test_gauss(self) -> None:
        """Test Gaussian filter on 2D and 3D arrays, including handling of NaNs and invalid input."""
        raster_array = get_array_and_mask(self.landsat_data)[0]
        raster_sm = gu.filters.gaussian_filter(raster_array, sigma=5)
        assert np.min(raster_array) <= np.min(raster_sm)
        assert np.max(raster_array) >= np.max(raster_sm)
        assert raster_array.shape == raster_sm.shape

        # Test with NaNs
        nan_count = 1000
        rng = np.random.default_rng(42)
        cols = rng.integers(0, high=self.landsat_data.width - 1, size=nan_count)
        rows = rng.integers(0, high=self.landsat_data.height - 1, size=nan_count)
        raster_with_nans = np.copy(self.landsat_data.data).squeeze()
        raster_with_nans[rows, cols] = np.nan

        raster_sm = gu.filters.gaussian_filter(raster_with_nans, sigma=10)
        assert np.nanmin(raster_with_nans) <= np.nanmin(raster_sm)
        assert np.nanmax(raster_with_nans) >= np.nanmax(raster_sm)

        # 3D arrays
        array_3d = np.vstack((raster_array[np.newaxis, :], raster_array[np.newaxis, :]))
        raster_sm = gu.filters.gaussian_filter(array_3d, sigma=5)
        assert array_3d.shape == raster_sm.shape

        # 1D array should raise
        data = raster_array[:, 0]
        pytest.raises(ValueError, gu.filters.gaussian_filter, data, sigma=5)


class TestStatisticalFilters:
    """Tests for statistical filters: mean, median, min, max."""

    landsat_data = gu.Raster(gu.examples.get_path("everest_landsat_b4")).astype(np.float32)

    @pytest.mark.parametrize(  # type: ignore
        "name, filter_func",
        [
            ("median", lambda arr: gu.filters.median_filter(arr, size=5)),
            ("mean", lambda arr: gu.filters.mean_filter(arr, size=5)),
            ("min", lambda arr: gu.filters.min_filter(arr, size=5)),
            ("max", lambda arr: gu.filters.max_filter(arr, size=5)),
        ],
    )
    def test_filters(self, name: str, filter_func: Callable[[NDArrayNum], NDArrayNum]) -> None:
        """Generic test for statistical filters with and without NaNs, including 3D and invalid 1D input."""
        raster_array = get_array_and_mask(self.landsat_data)[0]
        raster_filtered = filter_func(raster_array)

        rtol = 1e-5
        atol = 1e-5

        min_a, max_a = np.min(raster_array), np.max(raster_array)
        min_f, max_f = np.min(raster_filtered), np.max(raster_filtered)

        if name in ("median", "mean"):
            assert min_a <= min_f or np.isclose(min_a, min_f, rtol=rtol, atol=atol)
            assert max_a >= max_f or np.isclose(max_a, max_f, rtol=rtol, atol=atol)
        elif name == "min":
            assert np.isclose(min_a, min_f, rtol=rtol, atol=atol)
            assert max_a >= max_f or np.isclose(max_a, max_f, rtol=rtol, atol=atol)
        elif name == "max":
            assert min_a <= min_f or np.isclose(min_a, min_f, rtol=rtol, atol=atol)
            assert np.isclose(max_a, max_f, rtol=rtol, atol=atol)

        assert raster_array.shape == raster_filtered.shape

        # Test with NaNs
        nan_count = 1000
        rng = np.random.default_rng(42)
        cols = rng.integers(0, high=self.landsat_data.width - 1, size=nan_count)
        rows = rng.integers(0, high=self.landsat_data.height - 1, size=nan_count)
        raster_with_nans = np.copy(raster_array).squeeze()
        raster_with_nans[rows, cols] = np.nan

        raster_with_nans_filtered = filter_func(raster_with_nans)
        if name in ("median", "mean"):
            assert np.nanmin(raster_with_nans) <= np.nanmin(raster_with_nans_filtered)
            assert np.min(raster_filtered) == np.nanmin(raster_with_nans_filtered)
        elif name == "min":
            assert np.nanmin(raster_with_nans) == np.nanmin(raster_with_nans_filtered)
            assert np.min(raster_filtered) == np.nanmin(raster_with_nans_filtered)
            assert np.nanmax(raster_with_nans) >= np.nanmax(raster_with_nans_filtered)
        elif name == "max":
            assert np.nanmin(raster_with_nans) <= np.nanmin(raster_with_nans_filtered)
            assert np.nanmax(raster_with_nans) == np.nanmax(raster_with_nans_filtered)
            assert np.max(raster_filtered) == np.nanmax(raster_with_nans_filtered)

        if name != "mean":
            array_3d = np.vstack((raster_array[np.newaxis, :], raster_array[np.newaxis, :]))
            raster_filtered = filter_func(array_3d)
            assert array_3d.shape == raster_filtered.shape
            data = raster_array[:, 0]
            pytest.raises(ValueError, filter_func, data)

    def test_median_filter_nan_consistency(self) -> None:
        """Test that different median filter engines return consistent results with NaNs."""

        pytest.importorskip("numba")

        arr = np.array([[1, 2, np.nan], [4, np.nan, 6], [7, 8, 9]], dtype=np.float32)
        filtered_scipy = gu.filters.median_filter(arr, size=3, engine="scipy")
        filtered_numba = gu.filters.median_filter(arr, size=3, engine="numba")

        assert filtered_scipy.shape == arr.shape
        assert filtered_numba.shape == arr.shape
        assert np.allclose(filtered_scipy, filtered_numba, equal_nan=True)

    @pytest.mark.skipif(find_spec("numba") is not None, reason="Only runs if numba is missing.")  # type: ignore
    def test_filter_numba__missing_dep(self) -> None:
        """Test that when numba is missing, the proper import error is raised."""

        arr = np.array([[1, 2, np.nan], [4, np.nan, 6], [7, 8, 9]], dtype=np.float32)

        from geoutils.filters import median_filter

        with pytest.raises(ImportError, match="Optional dependency 'numba' required.*"):
            median_filter(arr, size=3, engine="numba")

    def test_median_filter_even_window_size_raises(self) -> None:
        """Ensure median filter raises with even window size."""
        arr = np.random.rand(10, 10).astype(np.float32)
        with pytest.raises(ValueError):
            gu.filters.median_filter(arr, size=4, engine="scipy")

    def test_min_max_filter_all_nans(self) -> None:
        """Test that min/max filters on all-NaN arrays return NaN arrays."""
        arr = np.full((5, 5), np.nan)
        filtered_min = gu.filters.min_filter(arr, size=3)
        filtered_max = gu.filters.max_filter(arr, size=3)
        assert np.all(np.isnan(filtered_min))
        assert np.all(np.isnan(filtered_max))


class TestDistanceFilter:
    """Tests for the distance-based outlier filter."""

    landsat_data = gu.Raster(gu.examples.get_path("everest_landsat_b4")).astype(np.float32)

    def test_dist_filter(self) -> None:
        """Check that distance filter removes outliers and preserves non-outliers."""
        landsat_data = self.landsat_data.copy()

        count = 1000
        rng = np.random.default_rng(42)
        cols = rng.integers(0, high=self.landsat_data.width - 1, size=count)
        rows = rng.integers(0, high=self.landsat_data.height - 1, size=count)
        landsat_data.data[rows, cols] = 5000

        filtered_landsat_data = gu.filters.distance_filter(landsat_data.data, radius=20, outlier_threshold=50)
        assert np.all(np.isnan(filtered_landsat_data[rows, cols]))
        assert landsat_data.data.shape == filtered_landsat_data.shape
        assert np.all(
            landsat_data.data[np.isfinite(filtered_landsat_data)]
            == filtered_landsat_data[np.isfinite(filtered_landsat_data)]
        )

        landsat_data.data[rows[:500], cols[:500]] = np.nan
        filtered_landsat_data = gu.filters.distance_filter(landsat_data.data, radius=20, outlier_threshold=50)
        assert np.all(np.isnan(filtered_landsat_data[rows, cols]))

    def test_distance_filter_all_nans(self) -> None:
        """Distance filter should return NaNs if all input is NaNs."""
        arr = np.full((10, 10), np.nan)
        filtered = gu.filters.distance_filter(arr, radius=2, outlier_threshold=1)
        assert np.all(np.isnan(filtered))

    def test_distance_filter_no_outliers(self) -> None:
        """Ensure no changes occur when no outliers are present."""
        arr = np.ones((10, 10)) * 10
        filtered = gu.filters.distance_filter(arr, radius=2, outlier_threshold=5)
        np.testing.assert_array_equal(arr, filtered)


class TestGenericFilter:
    """Tests for the generic filter API in GeoUtils."""

    landsat_data = gu.Raster(gu.examples.get_path("everest_landsat_b4")).astype(np.float32)

    def test_generic_filter(self) -> None:
        """Apply a scipy filter using the generic filter interface."""
        raster_array = get_array_and_mask(self.landsat_data)[0]
        raster_filtered = gu.filters.generic_filter(raster_array, scipy.ndimage.minimum_filter, size=5)
        scipy_filtered = scipy.ndimage.minimum_filter(raster_array, size=5)
        np.testing.assert_array_equal(raster_filtered, scipy_filtered)

    def test_generic_filter_1d_input_raises(self) -> None:
        """Generic filter should raise on 1D input."""
        arr = np.arange(10)
        with pytest.raises(ValueError):
            gu.filters.generic_filter(arr, scipy.ndimage.gaussian_filter, sigma=1)

    def test_filter_with_custom_callable(self) -> None:
        """Test using a custom function as filter method."""
        arr = np.arange(9).reshape(3, 3).astype(np.float32)

        def double(arr: NDArrayNum) -> NDArrayNum:
            return arr * 2

        filtered = gu.filters._filter(arr, method=double)
        np.testing.assert_array_equal(filtered, double(arr))

    def test_filter_with_invalid_method_type_raises(self) -> None:
        """Passing an invalid method type should raise a TypeError."""
        arr = np.arange(9).reshape(3, 3).astype(np.float32)
        with pytest.raises(ValueError):
            gu.filters._filter(arr, method="1234")


class TestRasterFilters:  # type: ignore
    """Tests for applying filters directly on `Raster` objects."""

    aster_dem_path = gu.examples.get_path("exploradores_aster_dem")

    def test_raster_filter_callable(self) -> None:
        """Apply a custom callable as a filter on a Raster object."""

        def double_filter(arr: NDArrayNum) -> NDArrayNum:
            return arr * 2

        data = np.array([[1, 1, 1, 1, 1], [2, 2, np.nan, 2, 2], [3, 3, 3, 3, 3], [4, 4, 4, np.nan, 4], [5, 5, 5, 5, 5]])
        transform = (30.0, 0.0, 478000.0, 0.0, -30.0, 3108140.0)
        raster = Raster.from_array(data, transform, 32645, np.nan)
        filtered = raster.filter(double_filter, inplace=False)
        expected_raster = raster.copy()
        expected_raster.data *= 2
        np.testing.assert_allclose(filtered.data.data, expected_raster.data.data)

    def test_raster_filter_inplace(self) -> None:
        """Check that in-place filtering modifies the original raster."""
        raster = gu.Raster(self.aster_dem_path)
        filtered_raster = raster.copy()
        filtered_raster.filter("gaussian", sigma=0, inplace=True)
        np.testing.assert_allclose(filtered_raster.data.data, raster.data.data, rtol=1e-1, atol=1e-1)

    def test_raster_filter_invalid(self) -> None:
        """Ensure invalid filter method raises appropriate exceptions."""
        raster = gu.Raster(self.aster_dem_path)
        with pytest.raises(ValueError, match="Unsupported filter method"):
            raster.filter("unknown_filter")
        with pytest.raises(TypeError, match="`method` must be a string or a callable"):
            raster.filter(12345, inplace=False)


class TestSyntheticsNansFilters:  # type: ignore
    """
    Test the filters with NaNs.
    """

    array = np.array([[1, 1, 1, 1, 1], [2, 2, np.nan, 2, 2], [3, 3, 3, 3, 3], [4, 4, 4, np.nan, 4], [5, 5, 5, 5, 5]])

    def test_min_filter_with_nan(self) -> None:
        """
        Test min filter with synthetics data containing NaNs
        """
        gt = np.array([[1, 1, 1, 1, 1], [1, 1, np.nan, 1, 1], [2, 2, 2, 2, 2], [3, 3, 3, np.nan, 3], [4, 4, 4, 4, 4]])
        test = gu.filters.min_filter(self.array, size=3)

        np.testing.assert_array_equal(gt, test)

    def test_max_filter_with_nan(self) -> None:
        """
        Test max filter with synthetics data containing NaNs
        """

        gt = np.array([[2, 2, 2, 2, 2], [3, 3, np.nan, 3, 3], [4, 4, 4, 4, 4], [5, 5, 5, np.nan, 5], [5, 5, 5, 5, 5]])
        test = gu.filters.max_filter(self.array, size=3)

        np.testing.assert_array_equal(gt, test)

    def test_mean_filter_with_nan(self) -> None:
        """
        Test mean filter with synthetics data containing NaNs
        """

        gt = np.array(
            [
                [1.5, 1.4, 1.4, 1.4, 1.5],
                [2, 2, np.nan, 2, 2],
                [3, 3.125, 3, 3.0, 2.8],
                [4, 4, 4, np.nan, 4],
                [4.5, 4.5, 4.6, 4.6, 4.66666667],
            ]
        )
        test = gu.filters.mean_filter(self.array, size=3)

        np.testing.assert_allclose(gt, test)

    def test_median_filter_with_nan(self) -> None:
        """
        Test median filter with synthetics data containing NaNs
        """

        gt = np.array(
            [
                [1.5, 1, 1, 1, 1.5],
                [2, 2, np.nan, 2, 2],
                [3, 3, 3, 3, 3],
                [4, 4, 4, np.nan, 4],
                [4.5, 4.5, 5, 5, 5],
            ]
        )
        test = gu.filters.median_filter(self.array, size=3)

        np.testing.assert_allclose(gt, test)

    def test_gaussian_filter_with_nan(self) -> None:
        """
        Test gaussian filter with synthetics data (from scipy documentation) containing NaNs
        """

        array = np.array(
            [
                [0.0, 2.0, 4.0, 6.0, 8.0],
                [10.0, 12.0, np.nan, 16.0, 18.0],
                [20.0, 22.0, 24.0, 26.0, 28.0],
                [30.0, 32.0, 34.0, np.nan, 38.0],
                [40.0, 42.0, 44.0, 46.0, 48.0],
            ]
        )

        test = gu.filters.gaussian_filter(array, sigma=1)

        # position (2,2)
        window = np.array([[12, np.nan, 16], [22, 24, 26], [32, 34, np.nan]])
        kernel = np.array([[1, np.nan, 1], [2, 4, 2], [1, 2, np.nan]])
        approx_gt = np.nansum(window * kernel) / np.nansum(kernel)
        np.testing.assert_allclose(approx_gt, test[2, 2], rtol=1e-1)

        # position (3, 1)
        window = np.array(
            [
                [20.0, 22.0, 24],
                [
                    30.0,
                    32.0,
                    34.0,
                ],
                [40.0, 42.0, 44.0],
            ]
        )
        kernel = np.array([[1, 2, 1], [2, 4, 2], [1, 2, 1]])
        approx_gt = np.nansum(window * kernel) / np.nansum(kernel)
        np.testing.assert_allclose(approx_gt, test[3, 1], rtol=1e-1)


@pytest.mark.parametrize(  # type: ignore
    "method, np_filter",
    [
        ("mean", np.nanmean),
        ("median", np.nanmedian),
        ("max", np.nanmax),
        ("min", np.nanmin),
    ],
)
def test_filter_against_center_value(method: str, np_filter: Callable[[NDArrayNum], NDArrayNum]) -> None:
    """
    Test filter against the center value of a function
    """
    arr = np.random.normal(size=(3, 3))

    arr_filtered = gu.filters._filter(arr, method=method, size=3)

    assert np.isclose(np_filter(arr), arr_filtered[1, 1], atol=1e-8)



"""Test the xdem.profiling functions."""

from __future__ import annotations

import glob
import os
import os.path as op
from importlib.util import find_spec

import pandas as pd
import pytest

import geoutils as gu
from geoutils import examples
from geoutils.profiler import Profiler


class TestProfiling:

    @pytest.mark.skipif(
        find_spec("psutil") is not None and find_spec("plotly") is not None,
        reason="Only runs if psutil or plotly is missing.",  # type: ignore
    )
    def test_profiling__missing_dep(self) -> None:
        """Test that the profiling functions fail with expected error when psutil is missing."""

        from geoutils.profiler import Profiler

        with pytest.raises(ImportError, match="Optional dependency 'plotly' required.*"):
            Profiler.enable()

    # Test that there's no crash when giving profiling configuration
    @pytest.mark.parametrize(
        "profiling_configuration",
        [(False, False, True), (True, False, True), (False, True, True), (True, True, True), (True, True, False)],
    )  # type: ignore
    @pytest.mark.parametrize("profiling_function", ["load", "get_stats", "subsample", "output_given"])  # type: ignore
    def test_profiling_configuration(
        self, profiling_configuration: tuple[bool, bool, bool], profiling_function: str, tmp_path: str
    ) -> None:
        """
        Test the all combinaisons of profiling with three examples of profiled functions.
        """
        pytest.importorskip("plotly")
        pytest.importorskip("psutil")

        s_gr = profiling_configuration[0]
        s_rd = profiling_configuration[1]
        output_given = profiling_configuration[2]

        Profiler.enable(save_graphs=s_gr, save_raw_data=s_rd)

        dem = gu.Raster(examples.get_path_test("everest_landsat_b4"))
        if profiling_function == "get_stats":
            dem.get_stats()
        if profiling_function == "subsample":
            gu.Raster.subsample(dem, 2)

        if output_given:
            Profiler.generate_summary(tmp_path)
            output_path = tmp_path
        else:
            os.chdir(tmp_path)
            Profiler.generate_summary()
            output_path = "output_profiling"

        # if profiling is activate
        if s_rd or s_gr:

            # in each case, output dir exist
            assert op.isdir(output_path)

            # if save_raw_data:
            if s_rd:
                # check pickle
                assert op.isfile(op.join(output_path, "raw_data.pickle"))

                # check data in pickle
                df = pd.read_pickle(op.join(output_path, "raw_data.pickle"))
                if profiling_function == "get_stats":
                    assert len(df) == 3
                elif profiling_function == "subsample":
                    assert len(df) == 2
                else:
                    assert len(df) == 1

            else:
                assert not op.isfile(op.join(output_path, "raw_data.pickle"))

            # if save_graphs:
            if s_gr:
                # check if all output graphs (time_graph + mem graph/profiled function called)
                # are generated
                assert op.isfile(op.join(output_path, "time_graph.html"))
                assert op.isfile(op.join(output_path, "memory_geoutils.raster.raster.__init__.html"))
                if profiling_function == "get_stats":
                    assert op.isfile(op.join(output_path, "memory_geoutils.stats.stats._statistics.html"))
                    assert op.isfile(op.join(output_path, "memory_geoutils.raster.raster.get_stats.html"))
                elif profiling_function == "sampling":
                    assert op.isfile(op.join(output_path, "memory_geoutils.raster.raster.subsample.html"))
            else:
                assert not len(glob.glob(op.join(output_path, "*.html")))

        else:
            # if profiling is deactivated : nothing generated in output dir
            assert not len(glob.glob(op.join(output_path, "*")))

    def test_profiling_functions_management(self) -> None:
        """
        Test the management of profiling functions information.
        """

        pytest.importorskip("plotly")
        pytest.importorskip("psutil")

        Profiler.enable(save_graphs=False, save_raw_data=True)

        assert len(Profiler.get_profiling_info()) == 0
        gu.Raster(examples.get_path_test("everest_landsat_b4"))

        assert len(Profiler.get_profiling_info()) == 1
        assert len(Profiler.get_profiling_info(function_name="geoutils.raster.raster.__init__")) == 1
        assert len(Profiler.get_profiling_info(function_name="geoutils.stats.stats.get_stats")) == 0
        assert len(Profiler.get_profiling_info(function_name="no_name")) == 0

        Profiler.reset()
        assert len(Profiler.get_profiling_info()) == 0

    def test_selections_functions(self) -> None:
        """
        Test the selection of functions to profile (all or by theirs names).
        """
        pytest.importorskip("plotly")
        pytest.importorskip("psutil")

        Profiler.enable(save_graphs=False, save_raw_data=True)
        Profiler.selection_functions(["geoutils.stats.stats._statistics"])
        dem = gu.Raster(examples.get_path_test("everest_landsat_b4"))

        dem.get_stats()
        assert len(Profiler.get_profiling_info()) == 1

        Profiler.selection_functions(["geoutils.raster.raster.__init__"])
        dem.get_stats()
        assert len(Profiler.get_profiling_info()) == 1

        Profiler.reset_selection_functions()
        dem.get_stats()
        assert len(Profiler.get_profiling_info()) == 3



"""
From/to Rasterio
================

This example demonstrates importing and exporting a :class:`rasterio.io.DatasetReader` or :class:`rasterio.io.DatasetReader` from and to a
:class:`~geoutils.Raster`.
"""

import rasterio as rio

# %%
# A raster can be imported from a :class:`rasterio.io.DatasetReader` or :class:`rasterio.io.MemoryFile` simply by instantiating :class:`~geoutils.Raster`.
import geoutils as gu

filename_rast = gu.examples.get_path("exploradores_aster_dem")
ds = rio.DatasetReader(filename_rast)
rast = gu.Raster(ds)
rast

# %%
# The data is unloaded, as when instantiated with a filename.
# The data will be loaded explicitly by any function requiring its :attr:`~geoutils.Raster.data`, such as :func:`~geoutils.Raster.show`.
rast.plot(cmap="terrain")

# %%
# We can also pass a :class:`rasterio.io.MemoryFile` during instantiation.

mem = rio.MemoryFile(open(filename_rast, "rb"))
rast = gu.Raster(mem)
rast

# %%
# The data is, as expected, already in memory.
#
# Finally, we can export a :class:`~geoutils.Raster` to a :class:`rasterio.io.DatasetReader` of a :class:`rasterio.io.MemoryFile` using
# :class:`~geoutils.Raster.to_rio_dataset`

rast.to_rio_dataset()



"""Test configuration file."""

import geoutils as gu


class TestConfig:
    def test_config_defaults(self) -> None:
        """Check defaults compared to file"""

        # Read file
        default_config = gu._config.GeoUtilsConfigDict()
        default_config._set_defaults(gu._config._config_ini_file)

        assert default_config == gu.config

    def test_config_set(self) -> None:
        """Check setting a non-default config argument by user"""

        # Default is True
        assert gu.config["shift_area_or_point"]

        # We set it to False and it should be updated
        gu.config["shift_area_or_point"] = False
        assert not gu.config["shift_area_or_point"]

        # Leave the test with the initial default
        gu.config["shift_area_or_point"] = True
        assert gu.config["shift_area_or_point"]

    def test_config_validator(self) -> None:
        """Check setting a config argument with a wrong input type converts it automatically"""

        # We input an "off" value, that should be converted to False
        gu.config["shift_area_or_point"] = "off"
        assert not gu.config["shift_area_or_point"]

        # Leave the test with initial default
        gu.config["shift_area_or_point"] = 1
        assert gu.config["shift_area_or_point"]



# Copyright (c) 2025 GeoUtils developers
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionalities at the interface of rasters and vectors."""

from __future__ import annotations

import warnings
from typing import Iterable, Literal

import affine
import geopandas as gpd
import numpy as np
import rasterio
import rasterio as rio
from rasterio import features, warp
from rasterio.crs import CRS
from rasterio.features import shapes

import geoutils as gu
from geoutils._misc import silence_rasterio_message
from geoutils._typing import NDArrayBool, NDArrayNum, Number
from geoutils.raster.georeferencing import _bounds


def _polygonize(
    source_raster: gu.Raster,
    target_values: Number | tuple[Number, Number] | list[Number] | NDArrayNum | Literal["all"],
    data_column_name: str,
) -> gu.Vector:
    """Polygonize a raster. See Raster.polygonize() for details."""

    # If target values is passed but does not correspond to 0 or 1, raise a warning
    if source_raster.is_mask:
        if target_values != "all" and (
            not isinstance(target_values, (int, np.integer, float, np.floating)) or target_values not in [0, 1]
        ):
            warnings.warn("Raster mask (boolean type) passed, using target value of 1 (True).")
        target_values = True

    # Mask a unique value set by a number
    if isinstance(target_values, (int, float, np.integer, np.floating)):
        if np.sum(source_raster.data == target_values) == 0:
            raise ValueError(f"no pixel with in_value {target_values}")

        bool_msk = np.array(source_raster.data == target_values).astype(np.uint8)

    # Mask values within boundaries set by a tuple
    elif isinstance(target_values, tuple):
        if np.sum((source_raster.data > target_values[0]) & (source_raster.data < target_values[1])) == 0:
            raise ValueError(f"no pixel with in_value between {target_values[0]} and {target_values[1]}")

        bool_msk = ((source_raster.data > target_values[0]) & (source_raster.data < target_values[1])).astype(np.uint8)

    # Mask specific values set by a sequence
    elif isinstance(target_values, list) or isinstance(target_values, np.ndarray):
        if np.sum(np.isin(source_raster.data, np.array(target_values))) == 0:
            raise ValueError("no pixel with in_value " + ", ".join(map("{}".format, target_values)))

        bool_msk = np.isin(source_raster.data, np.array(target_values)).astype("uint8")

    # Mask all valid values
    elif target_values == "all":
        # Using getmaskarray is necessary in case .data.mask is nomask (False)
        bool_msk = (~np.ma.getmaskarray(source_raster.data)).astype("uint8")

    else:
        raise ValueError("in_value must be a number, a tuple or a sequence")

    # GeoPandas.from_features() only supports certain dtypes, we find the best common dtype to optimize memory usage
    # TODO: this should be a function independent of polygonize, reused in several places
    gpd_dtypes = ["uint8", "uint16", "int16", "int32", "float32"]
    list_common_dtype_index = []
    for gpd_type in gpd_dtypes:
        polygonize_dtype = np.promote_types(gpd_type, source_raster.dtype)
        if str(polygonize_dtype) in gpd_dtypes:
            list_common_dtype_index.append(gpd_dtypes.index(gpd_type))
    if len(list_common_dtype_index) == 0:
        final_dtype = "float32"
    else:
        final_dtype_index = min(list_common_dtype_index)
        final_dtype = gpd_dtypes[final_dtype_index]

    results = (
        {"properties": {"raster_value": v}, "geometry": s}
        for i, (s, v) in enumerate(
            shapes(source_raster.data.astype(final_dtype), mask=bool_msk, transform=source_raster.transform)
        )
    )

    # Warning that comes from GeoPandas
    with silence_rasterio_message(param_name="MEM", warn_code="CPLE_AppDefined"):
        gdf = gpd.GeoDataFrame.from_features(list(results))
    gdf.insert(0, data_column_name, range(0, 0 + len(gdf)))
    gdf = gdf.set_geometry(col="geometry")
    gdf = gdf.set_crs(source_raster.crs)

    return gu.Vector(gdf)


def _rasterize(
    gdf: gpd.GeoDataFrame,
    raster: gu.Raster | None = None,
    crs: CRS | int | None = None,
    xres: float | None = None,
    yres: float | None = None,
    bounds: tuple[float, float, float, float] | None = None,
    in_value: int | float | Iterable[int | float] | None = None,
    out_value: int | float = 0,
) -> gu.Raster:
    if (raster is not None) and (crs is not None):
        raise ValueError("Only one of raster or crs can be provided.")

        # Reproject vector into requested CRS or rst CRS first, if needed
        # This has to be done first so that width/height calculated below are correct!
    if crs is None:
        crs = gdf.crs

    if raster is not None:
        crs = raster.crs  # type: ignore

    vect = gdf.to_crs(crs)

    # If no raster given, now use provided dimensions
    if raster is None:
        # At minimum, xres must be set
        if xres is None:
            raise ValueError("At least raster or xres must be set.")
        if yres is None:
            yres = xres

        # By default, use self's bounds
        if bounds is None:
            bounds = vect.total_bounds

        # Calculate raster shape
        left, bottom, right, top = bounds
        width = abs((right - left) / xres)
        height = abs((top - bottom) / yres)

        if width % 1 != 0 or height % 1 != 0:
            warnings.warn("Bounds not a multiple of xres/yres, use rounded bounds.")

        width = int(np.round(width))
        height = int(np.round(height))
        out_shape = (height, width)

        # Calculate raster transform
        transform = rio.transform.from_bounds(left, bottom, right, top, width, height)

    # otherwise use directly raster's dimensions
    else:
        out_shape = raster.shape  # type: ignore
        transform = raster.transform  # type: ignore

    # Set default burn value, index from 1 to len(self.ds)
    if in_value is None:
        in_value = gdf.index + 1

    # Rasterize geometry
    if isinstance(in_value, Iterable):
        if len(in_value) != len(vect.geometry):  # type: ignore
            raise ValueError(
                "in_value must have same length as self.ds.geometry, currently {} != {}".format(
                    len(in_value), len(vect.geometry)  # type: ignore
                )
            )

        out_geom = ((geom, value) for geom, value in zip(vect.geometry, in_value))

        mask = features.rasterize(shapes=out_geom, fill=out_value, out_shape=out_shape, transform=transform)

    elif isinstance(in_value, int | float | np.floating | np.integer):
        mask = features.rasterize(
            shapes=vect.geometry, fill=out_value, out_shape=out_shape, transform=transform, default_value=in_value
        )
    else:
        raise ValueError("in_value must be a single number or an iterable with same length as self.ds.geometry")

    output = gu.Raster.from_array(data=mask, transform=transform, crs=crs, nodata=None)

    return output


def _create_mask_pointcloud(gdf: gpd.GeoDataFrame, pts: gpd.GeoSeries) -> NDArrayBool:
    """Subfunction to create a point cloud mask using geopandas."""

    # Project to same CRS
    pts_reproj = pts.to_crs(crs=gdf.crs)

    # Check that points are contained no matter alignment
    contained = pts_reproj.within(gdf, align=False)

    # Extract resulting boolean array
    mask = contained.values

    return mask


def _create_mask_raster(
    gdf: gpd.GeoDataFrame, out_shape: tuple[int, int], transform: affine.Affine, crs: CRS
) -> NDArrayBool:
    """Subfunction to create a raster mask using rasterio.features.rasterize()."""

    # Copying GeoPandas dataframe before applying changes
    gdf = gdf.copy()

    # Crop vector geometries to avoid issues when reprojecting
    bounds = _bounds(transform=transform, shape=out_shape)
    left, bottom, right, top = bounds  # type: ignore
    x1, y1, x2, y2 = warp.transform_bounds(crs, gdf.crs, left, bottom, right, top)
    gdf = gdf.cx[x1:x2, y1:y2]

    # Reproject vector to raster CRS (almost always faster)
    gdf = gdf.to_crs(crs)

    # Rasterize geometry
    mask = features.rasterize(
        shapes=gdf.geometry, fill=0, out_shape=out_shape, transform=transform, default_value=1, dtype="uint8"
    ).astype("bool")

    return mask


def _create_mask(
    gdf: gpd.GeoDataFrame,
    ref: gu.Raster | gu.PointCloud | None = None,
    crs: CRS | None = None,
    res: float | tuple[float, float] | None = None,
    bounds: tuple[float, float, float, float] | None = None,
    points: tuple[NDArrayNum, NDArrayNum] | None = None,
) -> tuple[NDArrayBool, affine.Affine | None, gpd.GeoSeries | None, CRS]:
    """See Vector.create_mask for description."""

    # Raise errors for wrong inputs
    if ref is not None:
        if not isinstance(ref, (gu.Raster, gu.PointCloud)):
            raise ValueError("Reference must be a raster or a point cloud.")

    else:
        if res is None and points is None:
            raise ValueError(
                "Without a reference for masking, specify at least the resolution a raster mask, "
                "or the points coordinates for a point cloud mask."
            )

    # If raster reference or user-input exists, we compute a raster mask
    if (ref is not None and isinstance(ref, gu.Raster)) or res is not None:

        # For a reference, extract transform and CRS
        if ref is not None:
            transform = ref.transform
            out_shape = ref.shape
            crs = ref.crs

        # For a user-input res
        else:

            # By default, use self's CRS
            if crs is None:
                crs = gdf.crs

            # Case of a raster mask
            if res is not None:
                # Get resolution
                if isinstance(res, tuple):
                    xres, yres = res
                else:
                    xres = res
                    yres = res

            # Get bounds
            if bounds is None:
                bounds_shp = True
                bounds = gdf.total_bounds
            else:
                bounds_shp = False

            # Calculate raster shape
            left, bottom, right, top = bounds
            height = abs((right - left) / xres)
            width = abs((top - bottom) / yres)

            if width % 1 != 0 or height % 1 != 0:
                # Only warn if the bounds were provided, and not derived from the vector
                if not bounds_shp:
                    warnings.warn("Bounds not a multiple of resolution, using rounded bounds.")

            width = int(np.round(width))
            height = int(np.round(height))
            out_shape = (height, width)

            # Calculate raster transform
            transform = rio.transform.from_bounds(left, bottom, right, top, width, height)

        # Compute raster mask
        mask = _create_mask_raster(gdf=gdf, transform=transform, out_shape=out_shape, crs=crs)
        pts = None

    # For a point cloud reference or user-input, compute a point cloud mask
    else:

        # For a reference, extract geometry
        if ref is not None:
            pts = ref.ds.geometry

        else:

            if crs is None:
                crs = gdf.crs

            pts = gpd.points_from_xy(x=points[0], y=points[1], crs=crs)  # type: ignore

        mask = _create_mask_pointcloud(gdf=gdf, pts=pts)
        transform = None

    return mask, transform, crs, pts



"""Tests for raster-vector interfacing."""

from __future__ import annotations

import re
import warnings

import geopandas as gpd
import numpy as np
import pytest
from shapely import LineString, MultiLineString, MultiPolygon, Polygon

import geoutils as gu
from geoutils import examples


class TestRasterVectorInterface:

    # Create a synthetic vector file with a square of size 1, started at position (10, 10)
    poly1 = Polygon([(10, 10), (11, 10), (11, 11), (10, 11)])
    gdf = gpd.GeoDataFrame({"geometry": [poly1]}, crs="EPSG:4326")
    vector = gu.Vector(gdf)

    # Same with a square started at position (5, 5)
    poly2 = Polygon([(5, 5), (6, 5), (6, 6), (5, 6)])
    gdf = gpd.GeoDataFrame({"geometry": [poly2]}, crs="EPSG:4326")
    vector2 = gu.Vector(gdf)

    # Create a multipolygon with both
    multipoly = MultiPolygon([poly1, poly2])
    gdf = gpd.GeoDataFrame({"geometry": [multipoly]}, crs="EPSG:4326")
    vector_multipoly = gu.Vector(gdf)

    # Create a synthetic vector file with a square of size 5, started at position (8, 8)
    poly3 = Polygon([(8, 8), (13, 8), (13, 13), (8, 13)])
    gdf = gpd.GeoDataFrame({"geometry": [poly3]}, crs="EPSG:4326")
    vector_5 = gu.Vector(gdf)

    # Create a synthetic LineString geometry
    lines = LineString([(10, 10), (11, 10), (11, 11)])
    gdf = gpd.GeoDataFrame({"geometry": [lines]}, crs="EPSG:4326")
    vector_lines = gu.Vector(gdf)

    # Create a synthetic MultiLineString geometry
    multilines = MultiLineString([[(10, 10), (11, 10), (11, 11)], [(5, 5), (6, 5), (6, 6)]])
    gdf = gpd.GeoDataFrame({"geometry": [multilines]}, crs="EPSG:4326")
    vector_multilines = gu.Vector(gdf)

    def test_create_mask(self) -> None:
        """
        Test Vector.create_mask.
        """
        # First with given res and bounds -> Should be a 21 x 21 array with 0 everywhere except center pixel
        vector = self.vector.copy()
        out_mask = vector.create_mask(res=1, bounds=(0, 0, 21, 21), as_array=True)
        ref_mask = np.zeros((21, 21), dtype="bool")
        ref_mask[10, 10] = True
        assert out_mask.shape == (21, 21)
        assert np.all(ref_mask == out_mask)

        # Check that vector has not been modified by accident
        assert vector.bounds == self.vector.bounds
        assert len(vector.ds) == len(self.vector.ds)
        assert vector.crs == self.vector.crs

        # Then with a gu.Raster as reference, single band
        rst = gu.Raster.from_array(np.zeros((21, 21)), transform=(1.0, 0.0, 0.0, 0.0, -1.0, 21.0), crs="EPSG:4326")
        out_mask = vector.create_mask(rst, as_array=True)
        assert out_mask.shape == (21, 21)

        # With gu.Raster, 2 bands -> fails...
        # rst = gu.Raster.from_array(np.zeros((2, 21, 21)), transform=(1., 0., 0., 0., -1., 21.), crs='EPSG:4326')
        # out_mask = vector.create_mask(rst)

        # Check that no warning is raised when creating a mask with a xres not multiple of vector bounds
        mask = vector.create_mask(res=1.01)

        # Check that by default, create_mask returns a Mask
        assert isinstance(mask, gu.Raster) and mask.is_mask

        # Check that an error is raised if xres is not passed
        with pytest.raises(
            ValueError,
            match="Without a reference for masking, specify at least the resolution "
            "a raster mask, or the points coordinates for a point cloud mask.",
        ):
            vector.create_mask()

        # If the raster has the wrong type
        with pytest.raises(ValueError, match="Reference must be a raster or a point cloud."):
            vector.create_mask("lol")  # type: ignore

        # Check that a warning is raised if the bounds were passed specifically by the user
        with pytest.warns(UserWarning):
            vector.create_mask(res=1.01, bounds=(0, 0, 21, 21))

    landsat_b4_path = examples.get_path_test("everest_landsat_b4")
    landsat_b4_crop_path = gu.examples.get_path_test("everest_landsat_b4_cropped")
    everest_outlines_path = gu.examples.get_path_test("everest_rgi_outlines")
    aster_dem_path = gu.examples.get_path_test("exploradores_aster_dem")
    aster_outlines_path = gu.examples.get_path_test("exploradores_rgi_outlines")

    def test_rasterize(self) -> None:
        """Test rasterizing an EPSG:3426 dataset into a projection."""

        vct = gu.Vector(self.everest_outlines_path)
        rst = gu.Raster(self.landsat_b4_crop_path)

        # Use Web Mercator at 30 m.
        # Capture the warning on resolution not matching exactly bounds
        with pytest.warns(UserWarning):
            vct.rasterize(xres=30, crs=3857)

        # Typically, rasterize returns a raster
        burned_in2_out1 = vct.rasterize(raster=rst, in_value=2, out_value=1)
        assert isinstance(burned_in2_out1, gu.Raster)

        # For an in_value of 1 and out_value of 0 (default)
        burned_mask = vct.rasterize(raster=rst, in_value=1)
        assert isinstance(burned_mask, gu.Raster)
        # Convert to boolean
        burned_mask = burned_mask.astype(bool)

        # Check that rasterizing with in_value=1 is the same as creating a mask
        assert burned_mask.raster_equal(vct.create_mask(rst))

        # The two rasterization should match
        assert np.all(burned_in2_out1[burned_mask] == 2)
        assert np.all(burned_in2_out1[~burned_mask] == 1)

        # Check that errors are raised
        with pytest.raises(ValueError, match="Only one of raster or crs can be provided."):
            vct.rasterize(raster=rst, crs=3857)

    @pytest.mark.parametrize("example", [landsat_b4_path, aster_dem_path])  # type: ignore
    def test_polygonize(self, example: str) -> None:
        """Test that polygonize doesn't raise errors."""

        img = gu.Raster(example)

        # -- Test 1: basic functioning of polygonize --

        # Get unique value for image and the corresponding area
        value = np.unique(img)[0]
        pixel_area = np.count_nonzero(img.data == value) * img.res[0] * img.res[1]

        # Polygonize the raster for this value, and compute the total area
        polygonized = img.polygonize(target_values=value)
        polygon_area = polygonized.ds.area.sum()

        # Check that these two areas are approximately equal
        assert polygon_area == pytest.approx(pixel_area)
        assert isinstance(polygonized, gu.Vector)
        assert polygonized.crs == img.crs

        # Check default name of data column, and that defining a custom name works the same
        assert "id" in polygonized.ds.columns
        polygonized2 = img.polygonize(target_values=value, data_column_name="myname")
        assert "myname" in polygonized2.ds.columns
        assert np.array_equal(polygonized2.ds["myname"].values, polygonized.ds["id"].values)

        # -- Test 2: data types --

        # Check that polygonize works as expected for any input dtype (e.g. float64 being not supported by GeoPandas)
        for dtype in ["uint8", "int8", "uint16", "int16", "uint32", "int32", "float32", "float64"]:
            img_dtype = img.copy()
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore", category=UserWarning, message="dtype conversion will result in a " "loss of information.*"
                )
                warnings.filterwarnings(
                    "ignore",
                    category=UserWarning,
                    message="Unmasked values equal to the nodata value found in data array.*",
                )
                img_dtype = img_dtype.astype(dtype)
            value = np.unique(img_dtype)[0]
            img_dtype.polygonize(target_values=value)

        # And for a boolean object, such as a mask
        mask = img > value
        mask.polygonize(target_values=1)


class TestMaskVectorInterface:

    # Paths to example data
    landsat_b4_path = examples.get_path_test("everest_landsat_b4")
    landsat_rgb_path = examples.get_path_test("everest_landsat_rgb")
    everest_outlines_path = examples.get_path_test("everest_rgi_outlines")
    aster_dem_path = examples.get_path_test("exploradores_aster_dem")

    # Mask without nodata
    rst_landsat_b4 = gu.Raster(landsat_b4_path)
    mask_landsat_b4 = rst_landsat_b4 > np.nanmedian(rst_landsat_b4)
    # Mask with nodata
    rst_aster_dem = gu.Raster(aster_dem_path)
    mask_aster_dem = rst_aster_dem > np.nanmedian(rst_aster_dem)
    # Mask from an outline
    mask_everest = gu.Vector(everest_outlines_path).create_mask(gu.Raster(landsat_b4_path))

    @pytest.mark.parametrize("mask", [mask_landsat_b4, mask_aster_dem, mask_everest])  # type: ignore
    def test_polygonize(self, mask: gu.Raster) -> None:
        mask_orig = mask.copy()
        # Run default
        vect = mask.polygonize()
        # Check the dtype of the original mask was properly reconverted
        assert mask.data.dtype == bool
        # Check the original mask was not modified during polygonizing
        assert mask_orig.raster_equal(mask)

        # Check the output is cast into a vector
        assert isinstance(vect, gu.Vector)

        # Run with zero as target
        vect = mask.polygonize(target_values=0)
        assert isinstance(vect, gu.Vector)

        # Check a warning is raised when using a non-boolean value
        with pytest.warns(
            UserWarning, match=re.escape("Raster mask (boolean type) passed, using target value of 1 (" "True).")
        ):
            mask.polygonize(target_values=2)



"""
Crop a vector
=============

This example demonstrates the cropping of a vector using :func:`geoutils.Vector.crop`.
"""

# %%
# We open a raster and vector.

# sphinx_gallery_thumbnail_number = 3
import geoutils as gu

filename_rast = gu.examples.get_path("everest_landsat_b4_cropped")
filename_vect = gu.examples.get_path("everest_rgi_outlines")
rast = gu.Raster(filename_rast)
vect = gu.Vector(filename_vect)

# %%
# Let's plot the raster and vector. The raster has smaller extent than the vector.
rast.plot(cmap="Greys_r", alpha=0.7)
vect.plot(ref_crs=rast, fc="none", ec="tab:purple", lw=3)

# %%
# **First option:** using the raster as a reference to match, we crop the vector. We simply have to pass the :class:`~geoutils.Raster` as single argument to
# :func:`~geoutils.Vector.crop`. See :ref:`core-match-ref` for more details.

vect = vect.crop(rast)

# %%
# .. note::
#      By default, :func:`~geoutils.Vector.crop` is done in-place, replacing ``vect``. This behaviour can be modified by passing ``inplace=False``.
#

rast.plot(ax="new", cmap="Greys_r", alpha=0.7)
vect.plot(ref_crs=rast, fc="none", ec="tab:purple", lw=3)

# %%
# The :func:`~geoutils.Vector.crop` keeps all features with geometries intersecting the extent to crop to. We can also force a clipping of the geometries
# within the bounds using ``clip=True``.

vect = vect.crop(rast, clip=True)
rast.plot(ax="new", cmap="Greys_r", alpha=0.7)
vect.plot(ref_crs=rast, fc="none", ec="tab:purple", lw=3)

# %%
# **Second option:** we can pass other ``crop_geom`` argument to :func:`~geoutils.Vector.crop`, including another :class:`~geoutils.Vector` or a
# simple :class:`tuple` of bounds.

bounds = rast.get_bounds_projected(out_crs=vect.crs)
vect = vect.crop(crop_geom=(bounds.left + 0.5 * (bounds.right - bounds.left), bounds.bottom, bounds.right, bounds.top))

rast.plot(ax="new", cmap="Greys_r", alpha=0.7)
vect.plot(ref_crs=rast, fc="none", ec="tab:purple", lw=3)



####
# Part not shown in the doc, to get data files ready
import geoutils

landsat_b4_path = geoutils.examples.get_path("everest_landsat_b4")
landsat_b4_crop_path = geoutils.examples.get_path("everest_landsat_b4_cropped")
geoutils.Raster(landsat_b4_path).to_file("myraster1.tif")
geoutils.Raster(landsat_b4_crop_path).to_file("myraster2.tif")
####

import numpy as np
import rasterio as rio

# Opening of two rasters
rast1 = rio.io.DatasetReader("myraster1.tif")
rast2 = rio.io.DatasetReader("myraster2.tif")

# Equivalent of a match-reference reprojection
# (returns an array, not a raster-type object)
arr1_reproj, _ = rio.warp.reproject(
    source=rast1.read(),
    destination=np.ones(rast2.shape),
    src_transform=rast1.transform,
    src_crs=rast1.crs,
    src_nodata=rast1.nodata,
    dst_transform=rast2.transform,
    dst_crs=rast2.crs,
    dst_nodata=rast2.nodata,
)

# Equivalent of array interfacing
# (ensuring nodata and dtypes are rightly
# propagated through masked arrays)
ma1_reproj = np.ma.MaskedArray(data=arr1_reproj, mask=(arr1_reproj == rast2.nodata))
ma2 = rast2.read(masked=True)
ma_result = (1 + ma2) / (ma1_reproj)


# Equivalent of saving
# (requires to define a logical
# nodata for the data type)
def custom_func(dtype, nodata1, nodata2):
    return -9999


out_nodata = custom_func(dtype=ma_result.dtype, nodata1=rast1.nodata, nodata2=rast2.nodata)
with rio.open(
    "myresult.tif",
    mode="w",
    height=rast2.height,
    width=rast2.width,
    count=rast1.count,
    dtype=ma_result.dtype,
    crs=rast2.crs,
    transform=rast2.transform,
    nodata=rast2.nodata,
) as dst:
    dst.write(ma_result.filled(out_nodata))

####
# Part not shown in the docs, to clean up
import os

for file in ["myraster1.tif", "myraster2.tif", "myresult.tif"]:
    os.remove(file)
####



"""
Test file to quantify the impact of profiling and decorators.

- Profiling enabled and all decorators enabled
- Profiling enabled and all decorators disabled (empty selection functions)
- Profiling not called and decorators present in the code
- Profiling not called and decorators commented out in the code

The basic test run NTEST_GET_STATS get_stats with NTEST runs on exploradores_aster_dem. In terms of decorators involved,
there are geoutils.raster.raster.get_stats and geoutils.raster.raster.get_stats (so 2 per get_stats).

Package time and tracemalloc are used to compare duration and memory used for each subtests.
Each test saves its results in a CSV to reduce memory comsumption from dataframe operation.

WARNING: Take care of the last one that comments and decomments :
- @profiler.profile("geoutils.raster.raster.get_stats" ... decorator
- @profiler.profile("geoutils.stats.stats._statistics" ... decorator
WARNING: The last test only work with an Geoutils editable installation
WARNING: This benchmark work from benchmark directory
"""

import time
import tracemalloc

import pandas as pd  # type: ignore

import geoutils as gu
from geoutils.profiler import Profiler

NTEST = 20
NTEST_GET_STATS = 100


def replace(file: str, before: str, after: str):
    """
    Replace a string in a file by a specific string

    :param file path
    :param before string to replace
    :param before string to replace with
    """

    with open(file) as f:
        s = f.read()
        s = s.replace(before, after)
        with open(file, "w") as f:
            f.write(s)


def my_test(dem: gu.Raster):
    """
    Run the general test and extract its duration (s) and the max of memory used

    :param dem
    :return duration (s), peok_memory (KiB)
    """

    tracemalloc.reset_peak()
    tracemalloc.start()
    start_time = time.time()
    for t in range(NTEST_GET_STATS):
        dem.get_stats()
    duration = time.time() - start_time
    _, peak_memory = tracemalloc.get_traced_memory()
    return peak_memory / 1024, duration


def test_profiler_and_decorators_activated(dem: gu.Raster):
    """
    Run my_test NTEST time and compute duration (s) and memory consumption max
    with enabled Profiler and all decorated functions activated
    Save each test metrics in test_profiler_and_decorators_activated.csv
    """
    columns = ["profiling", "decorators", "memory", "time"]
    profiling_info = pd.DataFrame(columns=columns)

    for test in range(NTEST):
        Profiler.enable(save_graphs=True, save_raw_data=True)
        peak_memory, duration = my_test(dem)
        profiling_info.loc[len(profiling_info)] = {
            "profiling": "activated",
            "decorators": "presents and activated",
            "memory": peak_memory,
            "time": duration,
        }
    profiling_info.to_csv("test_profiler_and_decorators_activated.csv")


def test_profiler_and_decorators_not_selected(dem: gu.Raster):
    """
    Run my_test NTEST time and compute duration (s) and memory consumption max
    with enabled Profiler and all decorated functions not selected
    Save each test metrics in test_profiler_and_decorators_not_selected.csv
    """
    columns = ["profiling", "decorators", "memory", "time"]
    profiling_info = pd.DataFrame(columns=columns)

    for test in range(NTEST):
        Profiler.enable(save_graphs=True, save_raw_data=True)
        Profiler.selection_functions("my_test")
        peak_memory, duration = my_test(dem)
        profiling_info.loc[len(profiling_info)] = {
            "profiling": "activated",
            "decorators": "presents and desactivated",
            "memory": peak_memory,
            "time": duration,
        }
    profiling_info.to_csv("test_profiler_and_decorators_not_selected.csv")


def test_no_profiling_decorators_presents(dem: gu.Raster):
    """
    Run my_test NTEST time and compute duration (s) and memory consumption max
    with no Profiler and all decorated functions
    Save each test metrics in test_no_profiling_decorators_presents.csv
    """
    columns = ["profiling", "decorators", "memory", "time"]
    profiling_info = pd.DataFrame(columns=columns)

    for test in range(NTEST):
        peak_memory, duration = my_test(dem)
        profiling_info.loc[len(profiling_info)] = {
            "profiling": "desactivate",
            "decorators": "presents",
            "memory": peak_memory,
            "time": duration,
        }

    profiling_info.to_csv("test_no_profiling_decorators_presents.csv")


def test_no_profiling_no_decorators(dem: gu.Raster):
    """
    Run my_test NTEST time and compute duration (s) and memory consumption max
    with no Profiler and commented decorated functions
    Save each test metrics in test_no_profiling_no_decorators.csv
    """
    # remove get_stats decorator
    file_get_stats = "../geoutils/raster/raster.py"
    str_dec_get_stats = '    @profiler.profile("geoutils.raster.raster.get_stats"'
    str_no_dec_get_stats = '    #@profiler.profile("geoutils.raster.raster.get_stats"'
    replace(file_get_stats, str_dec_get_stats, str_no_dec_get_stats)

    # remove get_stats decorator
    file_stats = "../geoutils/stats/stats.py"
    str_dec_stats = '@profiler.profile("geoutils.stats.stats._statistics"'
    str_no_dec_stats = '#@profiler.profile("geoutils.stats.stats._statistics"'

    replace(file_stats, str_dec_stats, str_no_dec_stats)
    columns = ["profiling", "decorators", "memory", "time"]
    profiling_info = pd.DataFrame(columns=columns)

    for test in range(NTEST):
        peak_memory, duration = my_test(dem)
        profiling_info.loc[len(profiling_info)] = {
            "profiling": "desactivate",
            "decorators": "commented",
            "memory": peak_memory,
            "time": duration,
        }

    replace(file_get_stats, str_no_dec_get_stats, str_dec_get_stats)
    replace(file_stats, str_no_dec_stats, str_dec_stats)
    profiling_info.to_csv("test_no_profiling_no_decorators.csv")


dem = gu.Raster(gu.examples.get_path("exploradores_aster_dem"))

test_profiler_and_decorators_activated(dem)
test_profiler_and_decorators_not_selected(dem)
test_no_profiling_decorators_presents(dem)
test_no_profiling_no_decorators(dem)

df_profiler_and_decorators_activated = pd.read_csv("test_profiler_and_decorators_activated.csv", index_col=0)
df_profiler_and_decorators_not_selected = pd.read_csv("test_profiler_and_decorators_not_selected.csv", index_col=0)
df_no_profiling_decorators_presents = pd.read_csv("test_no_profiling_decorators_presents.csv", index_col=0)
df_no_profiling_no_decorators = pd.read_csv("test_no_profiling_no_decorators.csv", index_col=0)

df = pd.concat(
    [
        df_profiler_and_decorators_activated,
        df_profiler_and_decorators_not_selected,
        df_no_profiling_decorators_presents,
        df_no_profiling_no_decorators,
    ]
)

print("MEAN")
print(df.groupby(["profiling", "decorators"]).mean())
print("----------------")
print("STD")
print(df.groupby(["profiling", "decorators"]).std())



# Copyright (c) 2025 Centre National d'Etudes Spatiales (CNES).
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Contains profiling functions."""

import logging
import os
import time
import uuid
from functools import wraps
from multiprocessing import Pipe, connection
from threading import Thread
from typing import Any

import pandas as pd

from geoutils._misc import import_optional


class Profiler:
    """
    Main profiler class for Geoutils
    """

    enabled = False
    save_graphs = False
    save_raw_data = False
    columns = ["level", "uuid_function", "name", "uuid_parent", "time", "call_time", "memory"]
    _profiling_info = pd.DataFrame(columns=columns)
    selection_activated = False
    functions_selected = []
    running_processes = []

    @staticmethod
    def enable(save_graphs: bool = False, save_raw_data: bool = False) -> None:
        """
        Enables the profiler if save_graphs or save_raw_data is activated.

        :param save_graphs: Save the default graphs generated.
        :param save_raw_data: Save the raw data on calls as a .pickle file.
        """

        # To immediately raise errors if they are not installed
        import_optional("plotly")
        import_optional("psutil")

        Profiler.save_graphs = save_graphs
        Profiler.save_raw_data = save_raw_data
        Profiler.enabled = Profiler.save_graphs or Profiler.save_raw_data

        # Reset profiling information as a new Profiler is enabled
        Profiler.reset()

    @staticmethod
    def selection_functions(functions: list[str]) -> None:
        """
        List the functions to profile by their name

        :param functions: list of the functions name to profile
        """
        Profiler.selection_activated = True
        Profiler.functions_selected = functions

    @staticmethod
    def reset_selection_functions() -> None:
        """
        Cancel the possible selection of functions to profile
        """
        Profiler.selection_activated = False
        Profiler.functions_selected = []

    @staticmethod
    def add_profiling_info(info: dict[str, float | int | Any | str | list[Any] | None]) -> None:
        """
        Add profiling info to the profiling DataFrame.

        :param info: dictionary with profiling data keys
        """
        Profiler._profiling_info.loc[len(Profiler._profiling_info)] = {
            "level": info["level"],
            "uuid_function": info["uuid_function"],
            "name": info["name"],
            "uuid_parent": info["uuid_parent"],
            "time": info["time"],
            "call_time": info["call_time"],
            "memory": info["memory"],
        }

    @staticmethod
    def generate_summary(output: str = None) -> None:
        """
        Generate Profiling summary

        :param output: Output directory path, if None output is "output_profiling" in the current directory
        """

        import_optional("plotly")
        import plotly.express as px

        if output is None:
            output = "output_profiling"

        if not Profiler.enabled or len(Profiler._profiling_info) == 0:
            return

        if Profiler.save_raw_data or Profiler.save_graphs:
            os.makedirs(output, exist_ok=True)

        if Profiler.save_raw_data:
            Profiler._profiling_info.to_pickle(os.path.join(output, "raw_data.pickle"))
        Profiler._profiling_info["text_display"] = (
            Profiler._profiling_info["name"] + " (" + Profiler._profiling_info["time"].round(2).astype(str) + " s)"
        )

        if Profiler.save_graphs:
            # time profiling flame graph
            fig = px.icicle(
                Profiler._profiling_info,
                names="text_display",
                ids="uuid_function",
                parents="uuid_parent",
                values="time",
                title="Time profiling icicle graph (functions tagged only)",
                color="time",
                color_continuous_scale="thermal",
                branchvalues="total",
            )

            fig.update_traces(tiling_orientation="v")

            fig.write_html(os.path.join(output, "time_graph.html"))

            # memory profiling graph
            for _, call_row in Profiler._profiling_info[Profiler._profiling_info["memory"].notnull()].iterrows():
                path_fig = os.path.join(output, "memory_{}.html".format(call_row["name"]))
                Profiler.plot_trace_for_call(call_row["uuid_function"], "memory", path_fig)

    @staticmethod
    def get_profiling_info(function_name: str = None) -> pd.DataFrame:
        """
        Get profiling dataframe.
        If function_name is filled, it returns only matching rows (empty if no "name" matches).

        :param function_name: function name to show the profiled information
        :return dataframe information restrains function_name if filled
        """

        if Profiler._profiling_info.empty or not function_name:
            return Profiler._profiling_info

        if function_name:
            function_list = Profiler._profiling_info.loc[Profiler._profiling_info["name"] == function_name]
            return function_list

    @staticmethod
    def reset() -> None:
        """
        Reset profiling dataframe.
        """
        Profiler._profiling_info = pd.DataFrame(columns=Profiler.columns)

    @staticmethod
    def plot_trace_for_call(uuid_function: str, data_name: str, path_fig: str) -> None:
        """
        Plot memory (or any resource tracked) usage over time for a function call, with markers for its subcalls.
        :param uuid_function: UUID of the parent function call
        :param data_name: The name of the data to plot (if cpu consumption were to be added for example)
        :param path_fig: The path to save the output plot
        """

        import_optional("plotly")
        import plotly.graph_objects as go

        # Get the parent call entry
        parent_row = Profiler._profiling_info[Profiler._profiling_info["uuid_function"] == uuid_function]
        if parent_row.empty:
            return None
        parent_row = parent_row.iloc[0]

        call_start_time = parent_row["call_time"]
        times = [data[0] - call_start_time for data in parent_row[data_name]]
        values = [data[1] for data in parent_row[data_name]]

        # Collect subcalls (direct children)
        subcalls = Profiler._profiling_info[Profiler._profiling_info["uuid_parent"] == uuid_function]

        # Plot memory usage line
        fig = go.Figure()
        fig.add_trace(go.Scatter(x=times, y=values, mode="lines+markers", name=f"{data_name} usage"))

        # Base Y position for markers
        base_y = max(values)
        offset_step = (max(values) - min(values)) / 50  # how much higher each subsequent label goes
        current_offset = -offset_step * 2

        for _, row in list(subcalls.iterrows())[::-1]:
            sub_t = row["call_time"] - call_start_time
            sub_name = row["name"]

            # start function
            y_position = base_y + current_offset
            fig.add_trace(
                go.Scatter(
                    x=[sub_t],
                    y=[y_position],
                    mode="markers+text",
                    marker={
                        "color": "black",
                        "size": 8,
                    },
                    text=[sub_name],
                    textposition="middle right",  # text right next to marker at same height
                    showlegend=False,
                )
            )

            fig.add_shape(
                type="line",
                x0=sub_t,
                x1=sub_t,
                y0=min(values),
                y1=y_position,
                line={
                    "color": "black",
                    "width": 1,
                    "dash": "dot",
                },
            )

            # Increment offset for end function
            current_offset += offset_step
            y_position = base_y + current_offset

            fig.add_trace(
                go.Scatter(
                    x=[sub_t + row["time"]],
                    y=[y_position],
                    mode="markers+text",
                    marker={
                        "color": "black",
                        "size": 8,
                    },
                    text=["end " + sub_name],
                    textposition="middle right",  # text right next to marker at same height
                    showlegend=False,
                )
            )

            fig.add_shape(
                type="line",
                x0=sub_t + row["time"],
                x1=sub_t + row["time"],
                y0=min(values),
                y1=y_position,
                line={
                    "color": "black",
                    "width": 1,
                    "dash": "dot",
                },
            )

            # Increment offset for next marker
            current_offset += offset_step

        fig.update_layout(
            title="{} usage during {} call".format(data_name, parent_row["name"]),
            xaxis_title="Time (s)",
            yaxis_title="Memory (MB)",
            showlegend=True,
        )

        fig.write_html(path_fig)


def profile(name: str, interval: int | float = 0.005, memprof: bool = False):  # type: ignore
    """
    Geoutils profiling decorator

    To profile other functions and add them to the summary graphs and data, simply add the @profile decorator before
    them, providing a descriptive name. If you also want to track memory usage over time for a specific function call,
    set memprof=True in the decorator and if the function is too fast (or slow) for the default memory sampling
    interval, you can modify it with the interval parameter (in seconds).

    :param name: name of the function in the report
    :param interval: memory sampling interval (seconds)
    :param memprof: whether to profile the memory consumption

    :example:
        from geoutils import profiler

        @profiler.profile("my profiled function", memprof=True, interval=0.05)  # type: ignore
        def my_function():

    """

    def decorator_generator(func):  # type: ignore
        """
        Inner function
        """

        @wraps(func)
        def wrapper_profile(*args, **kwargs):  # type: ignore
            """
            Profiling wrapper

            Generate profiling logs of function, run

            :return: func(*args, **kwargs)
            """
            # if profiling is disabled, remove overhead

            if not Profiler.enabled:
                return func(*args, **kwargs)

            func_name = name
            if Profiler.selection_activated and name not in Profiler.functions_selected:
                return func(*args, **kwargs)

            uuid_function = str(uuid.uuid4())
            uuid_parent = Profiler.running_processes[-1] if Profiler.running_processes else "__main__"
            level = len(Profiler.running_processes)

            if name is None:
                func_name = func.__name__.capitalize()

            Profiler.running_processes.append(uuid_function)

            if memprof:
                # Launch memory profiling thread
                child_pipe, parent_pipe = Pipe()
                thread_monitoring = MemProf(os.getpid(), child_pipe, interval=interval)
                thread_monitoring.start()
                if parent_pipe.poll(1):  # wait for thread to start
                    parent_pipe.recv()

            start_time = time.time()
            res = func(*args, **kwargs)
            total_time = time.time() - start_time

            if memprof:
                # end memprofiling monitoring
                parent_pipe.send(0)

            Profiler.running_processes.pop(-1)  # remove function from call list

            func_data = {
                "level": level,
                "uuid_function": uuid_function,
                "name": func_name,
                "uuid_parent": uuid_parent,
                "time": total_time,
                "call_time": start_time,
                "memory": thread_monitoring.mem_data if memprof else None,
            }
            Profiler.add_profiling_info(func_data)
            return res

        return wrapper_profile

    return decorator_generator


class MemProf(Thread):
    """
    MemProf

    Profiling thread
    """

    def __init__(self, pid: int, pipe: connection.Connection, interval: float) -> None:
        """
        Init function of MemProf

        :param pid: The process ID of the monitored process
        :param pipe: The pipe used to send the end monitoring signal
        :param interval: Time interval (seconds) between memory measurements
        """
        psutil = import_optional("psutil")

        super().__init__()
        self.pipe = pipe
        self.interval = interval
        self.process = psutil.Process(pid)
        self.mem_data = []

    def run(self) -> None:
        """
        Run the memory profiling thread
        """

        try:
            # tell parent profiling is ready
            self.pipe.send(0)

            while True:

                timestamp = time.time()

                # Get memory in megabytes
                current_mem = self.process.memory_info().rss / 1000000
                self.mem_data.append((timestamp, current_mem))

                if self.pipe.poll(self.interval):
                    break

        except BrokenPipeError:
            logging.debug("Broken pipe error in log wrapper.")
